from pathlib import Path
import pcleaner.ocr.parsers as op
from tests.helpers import mock_file_path

from tests.mock_files import ocr_output


def test_parse_good_csv():
    file_path = mock_file_path("good_detected_text.csv", ocr_output)

    results, errors = op.parse_ocr_data(file_path)

    assert not errors
    assert len(results) == 1
    first_result = results[0]
    assert len(first_result.removed_box_data) == 2
    assert first_result.path == Path("img1.jpg")
    assert first_result.removed_box_data[0][2].area == 20_000


def test_parse_good_txt():
    file_path = mock_file_path("good_detected_text.txt", ocr_output)

    results, errors = op.parse_ocr_data(file_path)

    assert not errors
    assert len(results) == 2
    first_result = results[0]
    assert len(first_result.removed_box_data) == 2
    assert first_result.path == Path("page1.jpg")
    second_result = results[1]
    assert len(second_result.removed_box_data) == 3
    assert second_result.path == Path("page2.jpg")


def test_file_type_auto_detection_csv():
    file_path = mock_file_path("good_detected_text.csv", ocr_output)

    results, errors = op.parse_ocr_data(file_path)
    auto_results, auto_errors = op.parse_csv(file_path)

    assert results == auto_results
    assert errors == auto_errors


def test_file_type_auto_detection_txt():
    file_path = mock_file_path("good_detected_text.txt", ocr_output)

    results, errors = op.parse_ocr_data(file_path)
    auto_results, auto_errors = op.parse_plain_text(file_path)

    assert results == auto_results
    assert errors == auto_errors



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'OutputReview.ui'
##
## Created by: Qt User Interface Compiler version 6.7.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QAbstractItemView, QApplication, QComboBox, QFrame,
    QHBoxLayout, QLabel, QListView, QListWidget,
    QListWidgetItem, QPushButton, QSizePolicy, QSlider,
    QSpacerItem, QSplitter, QStackedWidget, QVBoxLayout,
    QWidget)

from pcleaner.gui.CustomQ.CElidedLabel import CElidedLabel
from pcleaner.gui.image_diff_viewers import (DifferenceViewer, OnionViewer, OverlayViewer, SwipeViewer)
from pcleaner.gui.image_viewer import ImageViewer

class Ui_OutputReview(object):
    def setupUi(self, OutputReview):
        if not OutputReview.objectName():
            OutputReview.setObjectName(u"OutputReview")
        OutputReview.resize(1400, 800)
        self.verticalLayout_7 = QVBoxLayout(OutputReview)
        self.verticalLayout_7.setObjectName(u"verticalLayout_7")
        self.verticalLayout_7.setContentsMargins(1, 1, 1, 1)
        self.splitter = QSplitter(OutputReview)
        self.splitter.setObjectName(u"splitter")
        self.splitter.setOrientation(Qt.Horizontal)
        self.layoutWidget = QWidget(self.splitter)
        self.layoutWidget.setObjectName(u"layoutWidget")
        self.verticalLayout_3 = QVBoxLayout(self.layoutWidget)
        self.verticalLayout_3.setObjectName(u"verticalLayout_3")
        self.verticalLayout_3.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout_2 = QHBoxLayout()
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.horizontalLayout_2.setContentsMargins(6, 6, 6, 7)
        self.label_image_count = QLabel(self.layoutWidget)
        self.label_image_count.setObjectName(u"label_image_count")
        self.label_image_count.setText(u"<image count>")

        self.horizontalLayout_2.addWidget(self.label_image_count)

        self.horizontalSpacer = QSpacerItem(0, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_2.addItem(self.horizontalSpacer)

        self.label = QLabel(self.layoutWidget)
        self.label.setObjectName(u"label")

        self.horizontalLayout_2.addWidget(self.label)

        self.horizontalSlider_icon_size = QSlider(self.layoutWidget)
        self.horizontalSlider_icon_size.setObjectName(u"horizontalSlider_icon_size")
        self.horizontalSlider_icon_size.setMinimum(1)
        self.horizontalSlider_icon_size.setMaximum(1000)
        self.horizontalSlider_icon_size.setOrientation(Qt.Horizontal)

        self.horizontalLayout_2.addWidget(self.horizontalSlider_icon_size)

        self.pushButton_prev = QPushButton(self.layoutWidget)
        self.pushButton_prev.setObjectName(u"pushButton_prev")
        self.pushButton_prev.setText(u"")
        icon = QIcon()
        iconThemeName = u"arrow-left"
        if QIcon.hasThemeIcon(iconThemeName):
            icon = QIcon.fromTheme(iconThemeName)
        else:
            icon.addFile(u".", QSize(), QIcon.Mode.Normal, QIcon.State.Off)

        self.pushButton_prev.setIcon(icon)
        self.pushButton_prev.setFlat(True)

        self.horizontalLayout_2.addWidget(self.pushButton_prev)

        self.pushButton_next = QPushButton(self.layoutWidget)
        self.pushButton_next.setObjectName(u"pushButton_next")
        self.pushButton_next.setText(u"")
        icon1 = QIcon()
        iconThemeName = u"arrow-right"
        if QIcon.hasThemeIcon(iconThemeName):
            icon1 = QIcon.fromTheme(iconThemeName)
        else:
            icon1.addFile(u".", QSize(), QIcon.Mode.Normal, QIcon.State.Off)

        self.pushButton_next.setIcon(icon1)
        self.pushButton_next.setFlat(True)

        self.horizontalLayout_2.addWidget(self.pushButton_next)

        self.horizontalLayout_2.setStretch(1, 1)
        self.horizontalLayout_2.setStretch(3, 3)

        self.verticalLayout_3.addLayout(self.horizontalLayout_2)

        self.image_list = QListWidget(self.layoutWidget)
        self.image_list.setObjectName(u"image_list")
        self.image_list.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOn)
        self.image_list.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.image_list.setEditTriggers(QAbstractItemView.NoEditTriggers)
        self.image_list.setProperty("showDropIndicator", False)
        self.image_list.setDragDropMode(QAbstractItemView.NoDragDrop)
        self.image_list.setTextElideMode(Qt.ElideLeft)
        self.image_list.setVerticalScrollMode(QAbstractItemView.ScrollPerPixel)
        self.image_list.setHorizontalScrollMode(QAbstractItemView.ScrollPerPixel)
        self.image_list.setMovement(QListView.Static)
        self.image_list.setFlow(QListView.LeftToRight)
        self.image_list.setProperty("isWrapping", True)
        self.image_list.setResizeMode(QListView.Adjust)
        self.image_list.setLayoutMode(QListView.Batched)
        self.image_list.setViewMode(QListView.IconMode)
        self.image_list.setUniformItemSizes(False)
        self.image_list.setBatchSize(50)
        self.image_list.setWordWrap(False)
        self.image_list.setSelectionRectVisible(True)

        self.verticalLayout_3.addWidget(self.image_list)

        self.splitter.addWidget(self.layoutWidget)
        self.layoutWidget1 = QWidget(self.splitter)
        self.layoutWidget1.setObjectName(u"layoutWidget1")
        self.verticalLayout = QVBoxLayout(self.layoutWidget1)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setSpacing(6)
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.horizontalLayout.setContentsMargins(6, 6, 6, 6)
        self.label_file_name = CElidedLabel(self.layoutWidget1)
        self.label_file_name.setObjectName(u"label_file_name")
        sizePolicy = QSizePolicy(QSizePolicy.Policy.MinimumExpanding, QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_file_name.sizePolicy().hasHeightForWidth())
        self.label_file_name.setSizePolicy(sizePolicy)
        self.label_file_name.setFrameShape(QFrame.NoFrame)
        self.label_file_name.setFrameShadow(QFrame.Raised)

        self.horizontalLayout.addWidget(self.label_file_name)

        self.pushButton_zoom_in = QPushButton(self.layoutWidget1)
        self.pushButton_zoom_in.setObjectName(u"pushButton_zoom_in")
        icon2 = QIcon()
        iconThemeName = u"zoom-in"
        if QIcon.hasThemeIcon(iconThemeName):
            icon2 = QIcon.fromTheme(iconThemeName)
        else:
            icon2.addFile(u".", QSize(), QIcon.Mode.Normal, QIcon.State.Off)

        self.pushButton_zoom_in.setIcon(icon2)
        self.pushButton_zoom_in.setFlat(True)

        self.horizontalLayout.addWidget(self.pushButton_zoom_in)

        self.pushButton_zoom_out = QPushButton(self.layoutWidget1)
        self.pushButton_zoom_out.setObjectName(u"pushButton_zoom_out")
        icon3 = QIcon()
        iconThemeName = u"zoom-out"
        if QIcon.hasThemeIcon(iconThemeName):
            icon3 = QIcon.fromTheme(iconThemeName)
        else:
            icon3.addFile(u".", QSize(), QIcon.Mode.Normal, QIcon.State.Off)

        self.pushButton_zoom_out.setIcon(icon3)
        self.pushButton_zoom_out.setFlat(True)

        self.horizontalLayout.addWidget(self.pushButton_zoom_out)

        self.pushButton_zoom_reset = QPushButton(self.layoutWidget1)
        self.pushButton_zoom_reset.setObjectName(u"pushButton_zoom_reset")
        icon4 = QIcon()
        iconThemeName = u"zoom-original"
        if QIcon.hasThemeIcon(iconThemeName):
            icon4 = QIcon.fromTheme(iconThemeName)
        else:
            icon4.addFile(u".", QSize(), QIcon.Mode.Normal, QIcon.State.Off)

        self.pushButton_zoom_reset.setIcon(icon4)
        self.pushButton_zoom_reset.setFlat(True)

        self.horizontalLayout.addWidget(self.pushButton_zoom_reset)

        self.pushButton_zoom_fit = QPushButton(self.layoutWidget1)
        self.pushButton_zoom_fit.setObjectName(u"pushButton_zoom_fit")
        icon5 = QIcon()
        iconThemeName = u"zoom-fit-best"
        if QIcon.hasThemeIcon(iconThemeName):
            icon5 = QIcon.fromTheme(iconThemeName)
        else:
            icon5.addFile(u".", QSize(), QIcon.Mode.Normal, QIcon.State.Off)

        self.pushButton_zoom_fit.setIcon(icon5)
        self.pushButton_zoom_fit.setFlat(True)

        self.horizontalLayout.addWidget(self.pushButton_zoom_fit)

        self.label_2 = QLabel(self.layoutWidget1)
        self.label_2.setObjectName(u"label_2")

        self.horizontalLayout.addWidget(self.label_2)

        self.comboBox_view_mode = QComboBox(self.layoutWidget1)
        self.comboBox_view_mode.addItem("")
        self.comboBox_view_mode.addItem("")
        self.comboBox_view_mode.addItem("")
        self.comboBox_view_mode.addItem("")
        self.comboBox_view_mode.addItem("")
        self.comboBox_view_mode.setObjectName(u"comboBox_view_mode")

        self.horizontalLayout.addWidget(self.comboBox_view_mode)

        self.pushButton_done = QPushButton(self.layoutWidget1)
        self.pushButton_done.setObjectName(u"pushButton_done")

        self.horizontalLayout.addWidget(self.pushButton_done)


        self.verticalLayout.addLayout(self.horizontalLayout)

        self.stackedWidget = QStackedWidget(self.layoutWidget1)
        self.stackedWidget.setObjectName(u"stackedWidget")
        self.page_side_by_side = QWidget()
        self.page_side_by_side.setObjectName(u"page_side_by_side")
        self.horizontalLayout_3 = QHBoxLayout(self.page_side_by_side)
        self.horizontalLayout_3.setSpacing(4)
        self.horizontalLayout_3.setObjectName(u"horizontalLayout_3")
        self.horizontalLayout_3.setContentsMargins(0, 0, 0, 0)
        self.image_viewer_sbs_master = ImageViewer(self.page_side_by_side)
        self.image_viewer_sbs_master.setObjectName(u"image_viewer_sbs_master")

        self.horizontalLayout_3.addWidget(self.image_viewer_sbs_master)

        self.image_viewer_sbs_slave = ImageViewer(self.page_side_by_side)
        self.image_viewer_sbs_slave.setObjectName(u"image_viewer_sbs_slave")
        self.image_viewer_sbs_slave.setEnabled(False)
        self.image_viewer_sbs_slave.setInteractive(False)

        self.horizontalLayout_3.addWidget(self.image_viewer_sbs_slave)

        self.horizontalLayout_3.setStretch(0, 1)
        self.horizontalLayout_3.setStretch(1, 1)
        self.stackedWidget.addWidget(self.page_side_by_side)
        self.page_swipe = QWidget()
        self.page_swipe.setObjectName(u"page_swipe")
        self.verticalLayout_4 = QVBoxLayout(self.page_swipe)
        self.verticalLayout_4.setSpacing(0)
        self.verticalLayout_4.setObjectName(u"verticalLayout_4")
        self.verticalLayout_4.setContentsMargins(0, 0, 0, 0)
        self.image_viewer_swipe = SwipeViewer(self.page_swipe)
        self.image_viewer_swipe.setObjectName(u"image_viewer_swipe")

        self.verticalLayout_4.addWidget(self.image_viewer_swipe)

        self.horizontalLayout_5 = QHBoxLayout()
        self.horizontalLayout_5.setObjectName(u"horizontalLayout_5")
        self.horizontalLayout_5.setContentsMargins(0, 6, 0, 6)
        self.horizontalSpacer_2 = QSpacerItem(0, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_5.addItem(self.horizontalSpacer_2)

        self.horizontalSlider_swipe = QSlider(self.page_swipe)
        self.horizontalSlider_swipe.setObjectName(u"horizontalSlider_swipe")
        sizePolicy1 = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.horizontalSlider_swipe.sizePolicy().hasHeightForWidth())
        self.horizontalSlider_swipe.setSizePolicy(sizePolicy1)
        self.horizontalSlider_swipe.setOrientation(Qt.Horizontal)

        self.horizontalLayout_5.addWidget(self.horizontalSlider_swipe)

        self.horizontalSpacer_3 = QSpacerItem(0, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_5.addItem(self.horizontalSpacer_3)

        self.horizontalLayout_5.setStretch(0, 1)
        self.horizontalLayout_5.setStretch(1, 3)
        self.horizontalLayout_5.setStretch(2, 1)

        self.verticalLayout_4.addLayout(self.horizontalLayout_5)

        self.stackedWidget.addWidget(self.page_swipe)
        self.page_onion = QWidget()
        self.page_onion.setObjectName(u"page_onion")
        self.verticalLayout_5 = QVBoxLayout(self.page_onion)
        self.verticalLayout_5.setSpacing(0)
        self.verticalLayout_5.setObjectName(u"verticalLayout_5")
        self.verticalLayout_5.setContentsMargins(0, 0, 0, 0)
        self.image_viewer_onion = OnionViewer(self.page_onion)
        self.image_viewer_onion.setObjectName(u"image_viewer_onion")

        self.verticalLayout_5.addWidget(self.image_viewer_onion)

        self.horizontalLayout_6 = QHBoxLayout()
        self.horizontalLayout_6.setObjectName(u"horizontalLayout_6")
        self.horizontalLayout_6.setContentsMargins(0, 6, 0, 6)
        self.horizontalSpacer_6 = QSpacerItem(0, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_6.addItem(self.horizontalSpacer_6)

        self.horizontalSlider_onion = QSlider(self.page_onion)
        self.horizontalSlider_onion.setObjectName(u"horizontalSlider_onion")
        sizePolicy1.setHeightForWidth(self.horizontalSlider_onion.sizePolicy().hasHeightForWidth())
        self.horizontalSlider_onion.setSizePolicy(sizePolicy1)
        self.horizontalSlider_onion.setOrientation(Qt.Horizontal)

        self.horizontalLayout_6.addWidget(self.horizontalSlider_onion)

        self.horizontalSpacer_7 = QSpacerItem(0, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_6.addItem(self.horizontalSpacer_7)

        self.horizontalLayout_6.setStretch(0, 1)
        self.horizontalLayout_6.setStretch(1, 3)
        self.horizontalLayout_6.setStretch(2, 1)

        self.verticalLayout_5.addLayout(self.horizontalLayout_6)

        self.stackedWidget.addWidget(self.page_onion)
        self.page_difference = QWidget()
        self.page_difference.setObjectName(u"page_difference")
        self.verticalLayout_6 = QVBoxLayout(self.page_difference)
        self.verticalLayout_6.setSpacing(0)
        self.verticalLayout_6.setObjectName(u"verticalLayout_6")
        self.verticalLayout_6.setContentsMargins(0, 0, 0, 0)
        self.image_viewer_difference = DifferenceViewer(self.page_difference)
        self.image_viewer_difference.setObjectName(u"image_viewer_difference")

        self.verticalLayout_6.addWidget(self.image_viewer_difference)

        self.horizontalLayout_7 = QHBoxLayout()
        self.horizontalLayout_7.setSpacing(0)
        self.horizontalLayout_7.setObjectName(u"horizontalLayout_7")
        self.horizontalLayout_7.setContentsMargins(0, 6, 0, 6)
        self.horizontalSpacer_8 = QSpacerItem(0, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_7.addItem(self.horizontalSpacer_8)

        self.horizontalSlider_difference = QSlider(self.page_difference)
        self.horizontalSlider_difference.setObjectName(u"horizontalSlider_difference")
        sizePolicy1.setHeightForWidth(self.horizontalSlider_difference.sizePolicy().hasHeightForWidth())
        self.horizontalSlider_difference.setSizePolicy(sizePolicy1)
        self.horizontalSlider_difference.setOrientation(Qt.Horizontal)

        self.horizontalLayout_7.addWidget(self.horizontalSlider_difference)

        self.horizontalSpacer_9 = QSpacerItem(0, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_7.addItem(self.horizontalSpacer_9)

        self.horizontalLayout_7.setStretch(0, 1)
        self.horizontalLayout_7.setStretch(1, 3)
        self.horizontalLayout_7.setStretch(2, 1)

        self.verticalLayout_6.addLayout(self.horizontalLayout_7)

        self.stackedWidget.addWidget(self.page_difference)
        self.page_overlay = QWidget()
        self.page_overlay.setObjectName(u"page_overlay")
        self.verticalLayout_2 = QVBoxLayout(self.page_overlay)
        self.verticalLayout_2.setSpacing(0)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.verticalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.image_viewer_overlay = OverlayViewer(self.page_overlay)
        self.image_viewer_overlay.setObjectName(u"image_viewer_overlay")

        self.verticalLayout_2.addWidget(self.image_viewer_overlay)

        self.horizontalLayout_8 = QHBoxLayout()
        self.horizontalLayout_8.setSpacing(0)
        self.horizontalLayout_8.setObjectName(u"horizontalLayout_8")
        self.horizontalLayout_8.setContentsMargins(0, 6, 0, 6)
        self.horizontalSpacer_10 = QSpacerItem(0, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_8.addItem(self.horizontalSpacer_10)

        self.horizontalSlider_overlay = QSlider(self.page_overlay)
        self.horizontalSlider_overlay.setObjectName(u"horizontalSlider_overlay")
        sizePolicy1.setHeightForWidth(self.horizontalSlider_overlay.sizePolicy().hasHeightForWidth())
        self.horizontalSlider_overlay.setSizePolicy(sizePolicy1)
        self.horizontalSlider_overlay.setOrientation(Qt.Horizontal)

        self.horizontalLayout_8.addWidget(self.horizontalSlider_overlay)

        self.horizontalSpacer_11 = QSpacerItem(0, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_8.addItem(self.horizontalSpacer_11)

        self.horizontalLayout_8.setStretch(0, 1)
        self.horizontalLayout_8.setStretch(1, 3)
        self.horizontalLayout_8.setStretch(2, 1)

        self.verticalLayout_2.addLayout(self.horizontalLayout_8)

        self.stackedWidget.addWidget(self.page_overlay)

        self.verticalLayout.addWidget(self.stackedWidget)

        self.splitter.addWidget(self.layoutWidget1)

        self.verticalLayout_7.addWidget(self.splitter)


        self.retranslateUi(OutputReview)

        self.stackedWidget.setCurrentIndex(0)


        QMetaObject.connectSlotsByName(OutputReview)
    # setupUi

    def retranslateUi(self, OutputReview):
        OutputReview.setWindowTitle(QCoreApplication.translate("OutputReview", u"Review", None))
        self.label.setText(QCoreApplication.translate("OutputReview", u"Icon Size:", None))
#if QT_CONFIG(tooltip)
        self.pushButton_prev.setToolTip(QCoreApplication.translate("OutputReview", u"Previous image", None))
#endif // QT_CONFIG(tooltip)
#if QT_CONFIG(tooltip)
        self.pushButton_next.setToolTip(QCoreApplication.translate("OutputReview", u"Next image", None))
#endif // QT_CONFIG(tooltip)
#if QT_CONFIG(tooltip)
        self.pushButton_zoom_in.setToolTip(QCoreApplication.translate("OutputReview", u"Zoom in", None))
#endif // QT_CONFIG(tooltip)
        self.pushButton_zoom_in.setText("")
#if QT_CONFIG(tooltip)
        self.pushButton_zoom_out.setToolTip(QCoreApplication.translate("OutputReview", u"Zoom out", None))
#endif // QT_CONFIG(tooltip)
        self.pushButton_zoom_out.setText("")
#if QT_CONFIG(tooltip)
        self.pushButton_zoom_reset.setToolTip(QCoreApplication.translate("OutputReview", u"Reset zoom", None))
#endif // QT_CONFIG(tooltip)
        self.pushButton_zoom_reset.setText("")
#if QT_CONFIG(tooltip)
        self.pushButton_zoom_fit.setToolTip(QCoreApplication.translate("OutputReview", u"Zoom to fit", None))
#endif // QT_CONFIG(tooltip)
        self.pushButton_zoom_fit.setText("")
        self.label_2.setText(QCoreApplication.translate("OutputReview", u"View Mode:", None))
        self.comboBox_view_mode.setItemText(0, QCoreApplication.translate("OutputReview", u"Side-by-Side", None))
        self.comboBox_view_mode.setItemText(1, QCoreApplication.translate("OutputReview", u"Swipe", None))
        self.comboBox_view_mode.setItemText(2, QCoreApplication.translate("OutputReview", u"Onion", None))
        self.comboBox_view_mode.setItemText(3, QCoreApplication.translate("OutputReview", u"Difference", None))
        self.comboBox_view_mode.setItemText(4, QCoreApplication.translate("OutputReview", u"Overlay", None))

        self.pushButton_done.setText(QCoreApplication.translate("OutputReview", u"Done", None))
    # retranslateUi




import math
from enum import IntEnum
from pathlib import Path

import PySide6.QtCore as Qc
import PySide6.QtGui as Qg
import PySide6.QtWidgets as Qw
from PySide6.QtCore import Slot, Signal
from loguru import logger
from natsort import natsorted

import pcleaner.config as cfg
import pcleaner.gui.gui_utils as gu
import pcleaner.gui.image_file as imf
import pcleaner.gui.image_viewer as iv
import pcleaner.gui.state_saver as ss
import pcleaner.gui.structures as gst
import pcleaner.output_structures as ost
from pcleaner.gui.ui_generated_files.ui_OutputReview import Ui_OutputReview
from pcleaner.helpers import f_plural


# The maximum size, will be smaller on one side if the image is not square.
THUMBNAIL_SIZE = 180


class ViewMode(IntEnum):
    SideBySide = 0
    Swipe = 1
    Onion = 2
    Difference = 3
    Overlay = 4


class OutputReviewWindow(Qw.QWidget, Ui_OutputReview):
    """
    A window to display the process results.
    """

    images: list[imf.ImageFile]
    target_output: ost.Output | None
    mask_outputs: list[ost.Output] | None
    show_isolated_text: bool
    config: cfg.Config
    confirm_closing: bool

    closed: Signal = Signal()

    first_slot_connection: bool
    first_sbs_slot_connection: bool

    shared_pixmap = gst.Shared[Qg.QPixmap]

    # Used so that the view will zoom to fit on the first load only.
    # This action needs to be delayed since the image dimensions are not known until the image is loaded,
    # since the size may change depending on the scale value.
    viewers_initialized: set

    min_thumbnail_size: int
    max_thumbnail_size: int

    state_saver: ss.StateSaver  # The state saver for the window.

    def __init__(
        self,
        parent=None,
        images: list[imf.ImageFile] = None,
        target_output: ost.Output | None = None,
        mask_outputs: list[ost.Output] | None = None,
        show_isolated_text: bool = False,
        config: cfg.Config = None,
        confirm_closing: bool = False,
    ):
        """
        Init the widget.

        :param parent: The parent widget.
        :param images: The images to display.
        :param target_output: The main output to display.
        :param mask_outputs: The mask outputs to display.
        :param show_isolated_text: Whether to show the isolated text instead of the output in
            certain view modes (side-by-side).
        :param config: The configuration object.
        :param confirm_closing: Whether to ask for confirmation before closing.
        """
        logger.info(f"Opening Review Window for {len(images)} outputs.")
        Qw.QWidget.__init__(self, parent)
        self.setupUi(self)
        self.setWindowFlag(Qc.Qt.Window)

        self.images = images
        self.target_output = target_output
        self.mask_outputs = mask_outputs
        self.show_isolated_text = show_isolated_text
        self.config = config
        self.confirm_closing = confirm_closing

        self.sort_images_by_path()

        self.first_load = set()
        self.first_slot_connection = True
        self.first_sbs_slot_connection = True
        self.shared_pixmap = gst.Shared[Qg.QPixmap | None]()

        self.calculate_thumbnail_size()
        self.init_arrow_buttons()
        self.init_image_list()
        self.init_image_viewers()
        self.image_list.currentItemChanged.connect(self.handle_image_change)
        self.comboBox_view_mode.currentIndexChanged.connect(self.handle_image_change)
        # Select the first image to start with.
        self.image_list.setCurrentRow(0)
        # Override Qt's dynamic scroll speed with a fixed, standard value.
        self.image_list.verticalScrollBar().setSingleStep(120)

        self.pushButton_done.clicked.connect(self.close)

        self.state_saver = ss.StateSaver("output_review")
        self.init_state_saver()
        self.state_saver.restore()

    def closeEvent(self, event: Qg.QCloseEvent) -> None:
        if self.confirm_closing:
            if (
                gu.show_question(
                    self,
                    self.tr("Finish Review"),
                    self.tr("Are you sure you want to finish the review?"),
                )
                == Qw.QMessageBox.Cancel
            ):
                event.ignore()
                return
        self.closed.emit()
        self.state_saver.save()
        event.accept()

    def cease_and_desist(self):
        """
        Just stop.
        """
        logger.info("Shutting up the review window before killing it.")
        self.confirm_closing = False
        self.closed.disconnect()
        self.close()

    def init_state_saver(self) -> None:
        """
        Load the state from the state saver.
        """
        self.state_saver.register(
            self,
            self.splitter,
            self.horizontalSlider_icon_size,
            self.comboBox_view_mode,
        )

    def sort_images_by_path(self) -> None:
        """
        Sort the images by their file path using natsort.
        This is necessary because the parallel batch processing will not preserve the order when
        many pictures are processed at once.
        """
        self.images = natsorted(self.images, key=lambda x: x.path)

    def calculate_thumbnail_size(self) -> None:
        """
        Use the current monitor's resolution to calculate the thumbnail size.
        The min value is 1% of the screen width, the max value is 100%.
        """
        screen = Qw.QApplication.primaryScreen()
        screen_size = screen.size()
        screen_width = screen_size.width()
        self.min_thumbnail_size = int(screen_width * 0.01)
        self.max_thumbnail_size = screen_width

    def to_log_scale(self, value: int) -> int:
        """
        Convert a linear value to a logarithmic scale.
        """
        min_log = math.log(self.min_thumbnail_size)
        max_log = math.log(self.max_thumbnail_size)
        scale = (max_log - min_log) / (
            self.horizontalSlider_icon_size.maximum() - self.horizontalSlider_icon_size.minimum()
        )
        log_value = min_log + (value - self.horizontalSlider_icon_size.minimum()) * scale
        return int(math.exp(log_value))

    def from_log_scale(self, value: int) -> int:
        """
        Convert a logarithmic value to a linear scale.
        """
        min_log = math.log(self.min_thumbnail_size)
        max_log = math.log(self.max_thumbnail_size)
        scale = (max_log - min_log) / (
            self.horizontalSlider_icon_size.maximum() - self.horizontalSlider_icon_size.minimum()
        )
        linear_value = (
            math.log(value) - min_log
        ) / scale + self.horizontalSlider_icon_size.minimum()
        return int(linear_value)

    def init_image_list(self) -> None:
        """
        Populate the image list with the images.
        """
        # Reduce the size of the list. Make it a 1/3 split.
        window_width = self.width()
        self.splitter.setSizes([window_width // 3, 2 * window_width // 3])

        label_text = f_plural(len(self.images), self.tr("image"), self.tr("images"))
        self.label_image_count.setText(f"{len(self.images)} {label_text}")

        for image in self.images:

            original_path = image.path
            output_path = image.outputs[self.target_output].path

            if output_path is None:
                logger.error(f"Output path for {image.path} is None.")

            try:
                self.image_list.addItem(
                    Qw.QListWidgetItem(Qg.QIcon(str(output_path)), str(original_path.stem))
                )
            except OSError:
                gu.show_exception(
                    self,
                    self.tr("Loading Error"),
                    self.tr("Failed to load image '{path}'").format(path=output_path),
                )

        # Use a logarithmic scale to distribute the values.
        self.horizontalSlider_icon_size.valueChanged.connect(self.update_icon_size)
        self.horizontalSlider_icon_size.setValue(self.from_log_scale(THUMBNAIL_SIZE))

        # Sanity check.
        if len(self.images) != self.image_list.count():
            logger.error("Failed to populate the image list correctly.")

    def init_arrow_buttons(self) -> None:
        """
        Connect the arrow buttons to the image list.
        """
        self.pushButton_next.clicked.connect(
            lambda: self.image_list.setCurrentRow(self.image_list.currentRow() + 1)
        )
        self.pushButton_prev.clicked.connect(
            lambda: self.image_list.setCurrentRow(self.image_list.currentRow() - 1)
        )

        # Disable the buttons if there is no next or previous image.
        self.image_list.currentRowChanged.connect(self.check_arrow_buttons)

    def check_arrow_buttons(self) -> None:
        """
        Check if the arrow buttons should be enabled or disabled.
        """
        current_row = self.image_list.currentRow()
        self.pushButton_prev.setEnabled(current_row > 0)
        self.pushButton_next.setEnabled(current_row < self.image_list.count() - 1)

    @Slot(int)
    def update_icon_size(self, size: int) -> None:
        """
        Update the icon size in the image list.

        :param size: The new size.
        """
        size = self.to_log_scale(size)
        self.image_list.setIconSize(Qc.QSize(size, size))
        for index, image in enumerate(self.images):
            item = self.image_list.item(index)
            text = image.path.name
            self.elide_text(item, text, int(size * 0.9))

    def elide_text(self, item, text, width):
        font_metrics = Qg.QFontMetrics(self.font())
        elided_text = font_metrics.elidedText(text, Qc.Qt.ElideLeft, width)
        item.setText(elided_text)

    def init_image_viewers(self) -> None:
        # Each one gets its own slider to manage the size of it themselves.
        self.image_viewer_swipe.connect_slider(self.horizontalSlider_swipe)
        self.image_viewer_onion.connect_slider(self.horizontalSlider_onion)
        self.image_viewer_difference.connect_slider(self.horizontalSlider_difference)
        self.image_viewer_overlay.connect_slider(self.horizontalSlider_overlay)
        # Create and share a common container with the viewers that support it.
        self.image_viewer_sbs_master.share_pixmap_container(self.shared_pixmap)
        self.image_viewer_swipe.share_pixmap_container(self.shared_pixmap)
        self.image_viewer_onion.share_pixmap_container(self.shared_pixmap)
        self.image_viewer_overlay.share_pixmap_container(self.shared_pixmap)

    @Slot(Qw.QListWidgetItem, Qw.QListWidgetItem)
    @Slot(Qw.QListWidgetItem)
    def handle_image_change(
        self, current: Qw.QListWidgetItem, previous: Qw.QListWidgetItem = None
    ) -> None:
        """
        Figure out what the new image is and display it.
        """
        # We don't need the previous image, so we ignore it.
        if current is None:
            return
        index = self.image_list.currentRow()
        image = self.images[index]
        self.switch_to_image(image)

    def current_view_mode(self) -> ViewMode:
        """
        Read the state of the combobox to determine the current view mode.

        :return: The current view mode.
        """
        return ViewMode(self.comboBox_view_mode.currentIndex())

    def switch_to_image(self, image: imf.ImageFile) -> None:
        """
        Show the image in the button in the image view.

        :param image: The image to show.
        """
        original_path = image.path
        output_path = image.outputs[self.target_output].path
        mask_paths = [image.outputs[mask].path for mask in self.mask_outputs]
        isolated_text_path = image.outputs[ost.Output.isolated_text].path

        self.label_file_name.setText(str(original_path))
        self.label_file_name.setToolTip(str(original_path))
        self.label_file_name.setElideMode(Qc.Qt.ElideLeft)

        view_mode = self.current_view_mode()
        # Invalidate the shared pixmap.
        self.shared_pixmap.set(None)
        self.stackedWidget.setCurrentIndex(view_mode)

        viewer_to_zoom: iv.ImageViewer | None = None

        match view_mode:
            case ViewMode.SideBySide:
                if not self.show_isolated_text:
                    self.load_side_by_side(original_path, output_path)
                else:
                    self.load_side_by_side(original_path, isolated_text_path)
                viewer_to_zoom = self.image_viewer_sbs_master
            case ViewMode.Swipe:
                self.load_swipe(original_path, output_path)
                viewer_to_zoom = self.image_viewer_swipe
            case ViewMode.Onion:
                self.load_onion(original_path, output_path)
                viewer_to_zoom = self.image_viewer_onion
            case ViewMode.Difference:
                self.load_difference(original_path, output_path)
                viewer_to_zoom = self.image_viewer_difference
            case ViewMode.Overlay:
                mask_color = self.config.current_profile.masker.debug_mask_color
                self.load_overlay(original_path, mask_paths, mask_color)
                viewer_to_zoom = self.image_viewer_overlay
            case _:
                logger.error("Invalid view mode.")

        # Disconnect prior slots if necessary.
        if not self.first_slot_connection:
            self.pushButton_zoom_in.clicked.disconnect()
            self.pushButton_zoom_out.clicked.disconnect()
            self.pushButton_zoom_fit.clicked.disconnect()
            self.pushButton_zoom_reset.clicked.disconnect()
        # Connect the zoom buttons to the image viewer.
        self.pushButton_zoom_in.clicked.connect(viewer_to_zoom.zoom_in)
        self.pushButton_zoom_out.clicked.connect(viewer_to_zoom.zoom_out)
        self.pushButton_zoom_fit.clicked.connect(viewer_to_zoom.zoom_fit)
        self.pushButton_zoom_reset.clicked.connect(viewer_to_zoom.zoom_reset)
        self.first_slot_connection = False

        # Set a delayed zoom to fit.
        if viewer_to_zoom not in self.first_load:
            Qc.QTimer.singleShot(0, viewer_to_zoom.zoom_fit)
            self.first_load.add(viewer_to_zoom)

    def load_side_by_side(self, original_path: Path, output_path: Path) -> None:
        """
        Load the images in the side-by-side view.
        This mode has 2 image viewers side by side.
        The master (left) shows the original image, the slave (right) shows the output image.
        Zoom events from the buttons are sent to the master, then the slave updates
        its zoom level to match the master.
        """
        # Check if the image wasn't already loaded.
        if self.image_viewer_sbs_master.loaded_image_path == original_path:
            return

        self.image_viewer_sbs_master.set_image(original_path)
        self.image_viewer_sbs_slave.set_image(output_path)

        # Sync master and slave.
        self.update_slave()

        # Connect the slave view to mimic the master at all times.
        if not self.first_sbs_slot_connection:
            return

        # The slave copies the master's scroll position and zoom transformation.
        self.image_viewer_sbs_master.zoom_factor_changed.connect(self.update_slave)
        self.image_viewer_sbs_master.horizontalScrollBar().valueChanged.connect(
            self.image_viewer_sbs_slave.horizontalScrollBar().setValue
        )
        self.image_viewer_sbs_master.verticalScrollBar().valueChanged.connect(
            self.image_viewer_sbs_slave.verticalScrollBar().setValue
        )
        self.first_sbs_slot_connection = False

    @Slot(Qc.QRectF)
    def update_slave(self) -> None:
        # Make sure the slave view stays consistent with the master.
        self.image_viewer_sbs_slave.set_transformation_mode(
            self.image_viewer_sbs_master.image_item.transformationMode()
        )
        self.image_viewer_sbs_slave.setRenderHints(self.image_viewer_sbs_master.renderHints())
        self.image_viewer_sbs_slave.setTransform(self.image_viewer_sbs_master.transform())

        # This may seem redundant, but without it the slave scroll bars jump wildly when
        # zooming in then out.
        # Likewise, the previous connection of the master's scroll positions to the
        # slave are necessary because zooming doesn't always trigger a scroll change.
        self.image_viewer_sbs_slave.verticalScrollBar().setValue(
            self.image_viewer_sbs_master.verticalScrollBar().value()
        )
        self.image_viewer_sbs_slave.horizontalScrollBar().setValue(
            self.image_viewer_sbs_master.horizontalScrollBar().value()
        )

    def load_swipe(self, original_path: Path, output_path: Path) -> None:
        """
        Load the images in the swipe view.
        This mode has 1 image viewer.
        The original image is shown on the left, the output image is shown on the right.
        The user can swipe the images horizontally to compare them.
        """
        if self.image_viewer_swipe.loaded_image_path == original_path:
            return
        self.image_viewer_swipe.set_images(original_path, output_path)

    def load_onion(self, original_path: Path, output_path: Path) -> None:
        """
        Load the images in the onion view.
        This mode has 1 image viewer.
        The output image is shown on top of the original image.
        The user can adjust the onion level to see more or less of the output image.
        """
        if self.image_viewer_onion.loaded_image_path == original_path:
            return
        self.image_viewer_onion.set_images(original_path, output_path)

    def load_difference(self, original_path: Path, output_path: Path) -> None:
        """
        Load the images in the difference view.
        This mode has 1 image viewer.
        The difference between the original and output images is shown.
        """
        if self.image_viewer_difference.loaded_image_path == original_path:
            return
        self.image_viewer_difference.set_images(original_path, output_path)

    def load_overlay(
        self, original_path: Path, mask_paths: list[Path], mask_color: tuple[int, int, int, int]
    ) -> None:
        """
        Load the images in the overlay view.
        This mode has 1 image viewer.
        The mask is shown on top of the original image.
        """
        if self.image_viewer_overlay.loaded_image_path == original_path:
            return
        self.image_viewer_overlay.set_images(original_path, mask_paths, mask_color)



import sys
import platform
import subprocess
from importlib import resources
from enum import Enum, auto
from pathlib import Path

import PySide6.QtWidgets as Qw
import xdg
from loguru import logger

import pcleaner.gui.gui_utils as gu
import pcleaner.data
from pcleaner.gui.ui_generated_files.ui_FileManagerIntegration import Ui_FileManagerExtension


class ExtensionTarget(Enum):
    Unsupported = auto()
    Dolphin = auto()
    WindowsExplorer = auto()


def get_extension_target() -> ExtensionTarget:
    """
    Get the file manager extension target.
    """
    if platform.system() == "Linux":
        # Make sure this isn't the flatpak version:
        if Path("/.flatpak-info").exists():
            return ExtensionTarget.Unsupported
        # Check which file manager is installed:
        try:
            subprocess.run(["which", "dolphin"], check=True)
            return ExtensionTarget.Dolphin
        except subprocess.CalledProcessError:
            return ExtensionTarget.Unsupported
    elif platform.system() == "Windows":
        # I've had it, fuck this. Someone else can try again.
        return ExtensionTarget.Unsupported
        # I dunno how to make this work well with the bare script version, so only exe supported.
        if getattr(sys, "frozen", False):
            return ExtensionTarget.WindowsExplorer
        else:
            return ExtensionTarget.Unsupported
    else:
        return ExtensionTarget.Unsupported


def get_plasma_service_menu_path() -> Path:
    """
    Get the default path, accounting for XDG.
    ~/.local/share/kio/servicemenus/panelcleaner.desktop

    :return: The path to store the service menu file as.
    """
    return Path(xdg.XDG_DATA_HOME) / "kio" / "servicemenus" / "panelcleaner.desktop"


class FileManagerExtension(Qw.QDialog, Ui_FileManagerExtension):
    """
    Allow the user to add or remove a file manager extension, if system is supported.
    """

    def __init__(
        self,
        parent=None,
    ) -> None:
        """
        Init the widget.

        :param parent: The parent widget.
        """
        Qw.QDialog.__init__(self, parent)
        self.setupUi(self)

        self.commandLinkButton_install.clicked.connect(self.install_extension)

        self.commandLinkButton_uninstall.clicked.connect(self.uninstall_extension)

        target = get_extension_target()
        if target == ExtensionTarget.Unsupported:
            gu.show_warning(
                self, self.tr("Unsupported system"), self.tr("This system is not supported.")
            )
            self.close()
        elif target == ExtensionTarget.Dolphin:
            self.commandLinkButton_install.setText(self.tr("Install for Dolphin"))
            self.commandLinkButton_uninstall.setText(self.tr("Uninstall for Dolphin"))
            self.commandLinkButton_install.setDescription(
                self.tr(
                    "The extension will be installed at {path}. "
                    "This doesn't require root permissions.".format(
                        path=get_plasma_service_menu_path()
                    )
                )
            )
            self.commandLinkButton_uninstall.setDescription(
                self.tr(
                    "The extension file will be deleted. This doesn't require root permissions."
                )
            )
        elif target == ExtensionTarget.WindowsExplorer:
            self.commandLinkButton_install.setText(self.tr("Install for Windows Explorer"))
            self.commandLinkButton_uninstall.setText(self.tr("Uninstall for Windows Explorer"))
            self.commandLinkButton_install.setDescription(
                self.tr(
                    "The extension will be installed in the Windows registry."
                    "This requires admin permissions."
                )
            )
            self.commandLinkButton_uninstall.setDescription(
                self.tr(
                    "The extension will be uninstalled from the Windows registry."
                    "This requires admin permissions."
                )
            )

    def install_extension(self) -> None:
        """
        Install the file manager extension.
        """
        target = get_extension_target()
        if target == ExtensionTarget.Unsupported:
            # This dialog shouldn't appear if the system is unsupported anyway.
            raise NotImplementedError("Unsupported system.")

        if target == ExtensionTarget.Dolphin:
            self.install_for_dolphin()
        elif target == ExtensionTarget.WindowsExplorer:
            self.install_for_windows_explorer()

    def uninstall_extension(self) -> None:
        """
        Uninstall the file manager extension.
        """
        target = get_extension_target()
        if target == ExtensionTarget.Unsupported:
            # This dialog shouldn't appear if the system is unsupported anyway.
            raise NotImplementedError("Unsupported system.")

        if target == ExtensionTarget.Dolphin:
            self.uninstall_for_dolphin()
        elif target == ExtensionTarget.WindowsExplorer:
            self.uninstall_for_windows_explorer()

    # ============================================== Install ==============================================

    def install_for_dolphin(self) -> None:
        """
        Install the file manager extension for Dolphin.
        """
        try:
            desktop_file_contents = format_desktop_entry()
            target_path = get_plasma_service_menu_path()

            msg = self.tr("Dolphin extension installed successfully.")
            if not target_path.parent.is_dir():
                msg = self.tr(
                    "Dolphin extension installed successfully.\n"
                    "You may need to log out and log back in (or restart) for the changes to take effect."
                )
            target_path.parent.mkdir(parents=True, exist_ok=True)

            if target_path.exists():
                target_path.unlink()

            target_path.write_text(desktop_file_contents)
            target_path.chmod(0o755)  # rwx r-x r-x Make it executable.
            gu.show_info(
                self,
                self.tr("Installation successful"),
                msg,
            )
        except Exception:
            gu.show_exception(
                self,
                self.tr("Installation failed"),
                self.tr("Failed to install Dolphin extension."),
            )

    def install_for_windows_explorer(self) -> None:
        """
        Install the file manager extension for Windows Explorer.
        """
        try:
            target_path = get_pcleaner_path()
            if target_path is None:
                logger.error("Can't determine where to point the windows explorer integration.")
                return
            logger.info(f"Pointing windows explorer integration to {target_path}")
            execution: list[Path | str] = get_executable_target()
            execution.extend(["create", target_path])
            logger.info(f"Executing subprocess: {execution}")
            subprocess.check_output(execution)
            gu.show_info(
                self, "Installation Success", "Successfully installed the Explorer integration."
            )
        except Exception:
            gu.show_exception(
                self,
                self.tr("Installation failed"),
                self.tr("Failed to install Windows Explorer extension."),
            )

    # ============================================== Uninstall ==============================================

    def uninstall_for_dolphin(self) -> None:
        """
        Uninstall the file manager extension for Dolphin.
        """
        try:
            target_path = get_plasma_service_menu_path()
            if target_path.exists():
                target_path.unlink()
                gu.show_info(
                    self,
                    self.tr("Uninstallation successful"),
                    self.tr("Dolphin extension uninstalled successfully."),
                )
            else:
                gu.show_info(
                    self,
                    self.tr("Uninstallation successful"),
                    self.tr("Dolphin extension was not installed."),
                )
        except Exception:
            gu.show_exception(
                self,
                self.tr("Uninstallation failed"),
                self.tr("Failed to uninstall Dolphin extension."),
            )

    def uninstall_for_windows_explorer(self) -> None:
        """
        Uninstall the file manager extension for Windows Explorer.
        """
        try:
            execution: list[Path | str] = get_executable_target()
            execution.extend(["delete", ""])
            logger.info(f"Executing subprocess: {execution}")
            subprocess.check_output(execution)
            gu.show_info(
                self, "Uninstallation Success", "Successfully uninstalled the Explorer integration."
            )
        except Exception:
            gu.show_exception(
                self,
                self.tr("Uninstallation failed"),
                self.tr("Failed to uninstall Windows Explorer extension."),
            )


def format_desktop_entry() -> str:
    """
    Format the desktop entry for the file manager extension.
    """
    return f"""\
[Desktop Entry]
Type=Service
MimeType=image/*;inode/directory;
Actions=cleanBoth;onlyMask;onlyClean;
X-KDE-Submenu=Panel Cleaner...


[Desktop Action cleanBoth]
Name=Save both the cleaned image and mask
Icon=background
Exec=pcleaner clean --notify %u
Icon=panelcleaner

[Desktop Action onlyMask]
Name=Save only the mask
Icon=background
Exec=pcleaner clean --notify -m %u
Icon=panelcleaner

[Desktop Action onlyClean]
Name=Save only the cleaned image
Icon=background
Exec=pcleaner clean --notify -c %u
Icon=panelcleaner

"""


def get_executable_target() -> list[Path]:
    """
    For a regular script, we want to return
    "path to python executable", "path to regedit script"

    But for the pyinstaller bundle, we can't use the bundled interpreter,
    so we need to instead point to the pre-built exe that bundles those two together.

    :return: execution path(s)
    """
    if getattr(sys, "frozen", False):
        # We need to fall back to the bundle.
        with resources.files(pcleaner.data) as data_path:
            exe = data_path / "WindowsExplorerIntegrationRegedit.exe"
        return [exe]

    else:
        python_interpreter = Path(sys.executable)
        with resources.files(pcleaner.data) as data_path:
            script_path = data_path / "windows_explorer_integration_regedit.py"
        return [python_interpreter, script_path]


def get_pcleaner_path() -> Path | None:
    # determine if application is a script file or frozen exe.
    if getattr(sys, "frozen", False):
        return Path(sys.executable)
    elif __file__:
        return Path(__file__).parent.parent / "main.py"
    return None



from functools import cache
from pathlib import Path

import pytesseract
from loguru import logger
from PIL import Image

import pcleaner.ocr.supported_languages as osl


@cache
def verify_tesseract():
    try:
        pytesseract.get_tesseract_version()
        return True
    except (pytesseract.TesseractNotFoundError, SystemExit) as e:
        logger.error(f"Tesseract verification failed: {e}")
        return False


@cache
def available_langs() -> set[osl.LanguageCode]:
    try:
        # We don't cover all languages, so abuse StrEnum string equality to compare
        # against membership before casting to LanguageCode.
        return set(
            osl.LanguageCode(code)
            for code in pytesseract.get_languages(config="")
            if code in osl.LanguageCode.__members__
        )
    except pytesseract.TesseractNotFoundError as e:
        logger.error(f"Error checking Tesseract available language data: {e}")
        return set()


def load_image(img_or_path) -> Image.Image:
    if isinstance(img_or_path, (str, Path)):
        return Image.open(img_or_path)
    elif isinstance(img_or_path, Image.Image):
        return img_or_path
    else:
        raise ValueError(f"img_or_path must be a path or PIL.Image, got: {type(img_or_path)}")


def cleanup_text(text: str) -> str:
    """
    Strip newlines and multiple spaces.
    """
    return text.replace("\n", " ").replace("\r", " ").replace("\t", " ").replace("\f", " ").strip()


class TesseractOcr:
    def __init__(self, lang: str | None = None):
        self.lang = lang
        # self.config = r'--psm 4 -c preserve_interword_spaces=1'
        # self.config = r'--psm 4'
        self.config = r""

    @staticmethod
    def is_tesseract_available() -> bool:
        return verify_tesseract()

    @staticmethod
    def langs() -> set[osl.LanguageCode]:
        return available_langs()

    def __call__(
        self,
        img_or_path: Image.Image | Path | str,
        lang: str | None = None,
        config: str | None = None,
        **kwargs,
    ) -> str:
        if not self.is_tesseract_available():
            raise RuntimeError("Tesseract OCR is not installed or not found.")
        if lang and lang not in self.langs():
            raise RuntimeError(f"Tesseract OCR language pack '{lang}' not found.")
        img = load_image(img_or_path)
        try:
            raw_text = pytesseract.image_to_string(
                img, lang=lang or self.lang, config=config or self.config
            )
        except Exception:
            raw_text = ""
        if not raw_text:  # try again with sparse text
            try:
                d = pytesseract.image_to_data(
                    img,
                    lang=lang or self.lang,
                    config=r"--psm 11",
                    output_type=pytesseract.Output.DICT,
                )
                raw_text = " ".join(_ for _ in d.get("text", ()) if _)
            except Exception as e:
                logger.error(f"Failed to run Tesseract with --psm 11: {e}")
        return cleanup_text(raw_text)



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'OcrReview.ui'
##
## Created by: Qt User Interface Compiler version 6.7.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QAbstractItemView, QApplication, QComboBox, QDialog,
    QFrame, QHBoxLayout, QHeaderView, QLabel,
    QListView, QListWidget, QListWidgetItem, QPushButton,
    QSizePolicy, QSlider, QSpacerItem, QSplitter,
    QTableWidgetItem, QVBoxLayout, QWidget)

from pcleaner.gui.CustomQ.CComboBox import CComboBox
from pcleaner.gui.CustomQ.CElidedLabel import CElidedLabel
from pcleaner.gui.CustomQ.CTableWidget import CTableWidget
from pcleaner.gui.image_viewer import BubbleImageViewer

class Ui_OcrReview(object):
    def setupUi(self, OcrReview):
        if not OcrReview.objectName():
            OcrReview.setObjectName(u"OcrReview")
        OcrReview.resize(1400, 800)
        OcrReview.setWindowTitle(u"Review")
        OcrReview.setModal(True)
        self.verticalLayout_2 = QVBoxLayout(OcrReview)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.verticalLayout_2.setContentsMargins(1, 1, 1, 1)
        self.splitter = QSplitter(OcrReview)
        self.splitter.setObjectName(u"splitter")
        self.splitter.setOrientation(Qt.Horizontal)
        self.layoutWidget1 = QWidget(self.splitter)
        self.layoutWidget1.setObjectName(u"layoutWidget1")
        self.verticalLayout_3 = QVBoxLayout(self.layoutWidget1)
        self.verticalLayout_3.setSpacing(6)
        self.verticalLayout_3.setObjectName(u"verticalLayout_3")
        self.verticalLayout_3.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout_2 = QHBoxLayout()
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.horizontalLayout_2.setContentsMargins(6, 6, 6, 7)
        self.label_image_count = QLabel(self.layoutWidget1)
        self.label_image_count.setObjectName(u"label_image_count")
        self.label_image_count.setText(u"<image count>")

        self.horizontalLayout_2.addWidget(self.label_image_count)

        self.horizontalSpacer = QSpacerItem(0, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_2.addItem(self.horizontalSpacer)

        self.label = QLabel(self.layoutWidget1)
        self.label.setObjectName(u"label")

        self.horizontalLayout_2.addWidget(self.label)

        self.horizontalSlider_icon_size = QSlider(self.layoutWidget1)
        self.horizontalSlider_icon_size.setObjectName(u"horizontalSlider_icon_size")
        self.horizontalSlider_icon_size.setMinimum(1)
        self.horizontalSlider_icon_size.setMaximum(1000)
        self.horizontalSlider_icon_size.setOrientation(Qt.Horizontal)

        self.horizontalLayout_2.addWidget(self.horizontalSlider_icon_size)

        self.pushButton_prev = QPushButton(self.layoutWidget1)
        self.pushButton_prev.setObjectName(u"pushButton_prev")
        self.pushButton_prev.setText(u"")
        icon = QIcon()
        iconThemeName = u"arrow-left"
        if QIcon.hasThemeIcon(iconThemeName):
            icon = QIcon.fromTheme(iconThemeName)
        else:
            icon.addFile(u".", QSize(), QIcon.Mode.Normal, QIcon.State.Off)

        self.pushButton_prev.setIcon(icon)
        self.pushButton_prev.setFlat(True)

        self.horizontalLayout_2.addWidget(self.pushButton_prev)

        self.pushButton_next = QPushButton(self.layoutWidget1)
        self.pushButton_next.setObjectName(u"pushButton_next")
        self.pushButton_next.setText(u"")
        icon1 = QIcon()
        iconThemeName = u"arrow-right"
        if QIcon.hasThemeIcon(iconThemeName):
            icon1 = QIcon.fromTheme(iconThemeName)
        else:
            icon1.addFile(u".", QSize(), QIcon.Mode.Normal, QIcon.State.Off)

        self.pushButton_next.setIcon(icon1)
        self.pushButton_next.setFlat(True)

        self.horizontalLayout_2.addWidget(self.pushButton_next)

        self.horizontalLayout_2.setStretch(1, 1)
        self.horizontalLayout_2.setStretch(3, 3)

        self.verticalLayout_3.addLayout(self.horizontalLayout_2)

        self.image_list = QListWidget(self.layoutWidget1)
        self.image_list.setObjectName(u"image_list")
        self.image_list.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOn)
        self.image_list.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.image_list.setEditTriggers(QAbstractItemView.NoEditTriggers)
        self.image_list.setProperty("showDropIndicator", False)
        self.image_list.setDragDropMode(QAbstractItemView.NoDragDrop)
        self.image_list.setTextElideMode(Qt.ElideLeft)
        self.image_list.setVerticalScrollMode(QAbstractItemView.ScrollPerPixel)
        self.image_list.setHorizontalScrollMode(QAbstractItemView.ScrollPerPixel)
        self.image_list.setMovement(QListView.Static)
        self.image_list.setFlow(QListView.LeftToRight)
        self.image_list.setProperty("isWrapping", True)
        self.image_list.setResizeMode(QListView.Adjust)
        self.image_list.setLayoutMode(QListView.Batched)
        self.image_list.setViewMode(QListView.IconMode)
        self.image_list.setUniformItemSizes(False)
        self.image_list.setBatchSize(50)
        self.image_list.setWordWrap(False)
        self.image_list.setSelectionRectVisible(True)

        self.verticalLayout_3.addWidget(self.image_list)

        self.splitter.addWidget(self.layoutWidget1)
        self.layoutWidget = QWidget(self.splitter)
        self.layoutWidget.setObjectName(u"layoutWidget")
        self.verticalLayout = QVBoxLayout(self.layoutWidget)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setSpacing(6)
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.horizontalLayout.setContentsMargins(6, 6, 6, 6)
        self.label_file_name = CElidedLabel(self.layoutWidget)
        self.label_file_name.setObjectName(u"label_file_name")
        sizePolicy = QSizePolicy(QSizePolicy.Policy.MinimumExpanding, QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_file_name.sizePolicy().hasHeightForWidth())
        self.label_file_name.setSizePolicy(sizePolicy)
        self.label_file_name.setFrameShape(QFrame.NoFrame)
        self.label_file_name.setFrameShadow(QFrame.Raised)

        self.horizontalLayout.addWidget(self.label_file_name)

        self.pushButton_zoom_in = QPushButton(self.layoutWidget)
        self.pushButton_zoom_in.setObjectName(u"pushButton_zoom_in")
        icon2 = QIcon()
        iconThemeName = u"zoom-in"
        if QIcon.hasThemeIcon(iconThemeName):
            icon2 = QIcon.fromTheme(iconThemeName)
        else:
            icon2.addFile(u".", QSize(), QIcon.Mode.Normal, QIcon.State.Off)

        self.pushButton_zoom_in.setIcon(icon2)
        self.pushButton_zoom_in.setFlat(True)

        self.horizontalLayout.addWidget(self.pushButton_zoom_in)

        self.pushButton_zoom_out = QPushButton(self.layoutWidget)
        self.pushButton_zoom_out.setObjectName(u"pushButton_zoom_out")
        icon3 = QIcon()
        iconThemeName = u"zoom-out"
        if QIcon.hasThemeIcon(iconThemeName):
            icon3 = QIcon.fromTheme(iconThemeName)
        else:
            icon3.addFile(u".", QSize(), QIcon.Mode.Normal, QIcon.State.Off)

        self.pushButton_zoom_out.setIcon(icon3)
        self.pushButton_zoom_out.setFlat(True)

        self.horizontalLayout.addWidget(self.pushButton_zoom_out)

        self.pushButton_zoom_reset = QPushButton(self.layoutWidget)
        self.pushButton_zoom_reset.setObjectName(u"pushButton_zoom_reset")
        icon4 = QIcon()
        iconThemeName = u"zoom-original"
        if QIcon.hasThemeIcon(iconThemeName):
            icon4 = QIcon.fromTheme(iconThemeName)
        else:
            icon4.addFile(u".", QSize(), QIcon.Mode.Normal, QIcon.State.Off)

        self.pushButton_zoom_reset.setIcon(icon4)
        self.pushButton_zoom_reset.setFlat(True)

        self.horizontalLayout.addWidget(self.pushButton_zoom_reset)

        self.pushButton_zoom_fit = QPushButton(self.layoutWidget)
        self.pushButton_zoom_fit.setObjectName(u"pushButton_zoom_fit")
        icon5 = QIcon()
        iconThemeName = u"zoom-fit-best"
        if QIcon.hasThemeIcon(iconThemeName):
            icon5 = QIcon.fromTheme(iconThemeName)
        else:
            icon5.addFile(u".", QSize(), QIcon.Mode.Normal, QIcon.State.Off)

        self.pushButton_zoom_fit.setIcon(icon5)
        self.pushButton_zoom_fit.setFlat(True)

        self.horizontalLayout.addWidget(self.pushButton_zoom_fit)

        self.comboBox_view_mode = QComboBox(self.layoutWidget)
        self.comboBox_view_mode.addItem("")
        self.comboBox_view_mode.addItem("")
        self.comboBox_view_mode.setObjectName(u"comboBox_view_mode")

        self.horizontalLayout.addWidget(self.comboBox_view_mode)

        self.line = QFrame(self.layoutWidget)
        self.line.setObjectName(u"line")
        self.line.setFrameShape(QFrame.Shape.VLine)
        self.line.setFrameShadow(QFrame.Shadow.Sunken)

        self.horizontalLayout.addWidget(self.line)

        self.pushButton_row_up = QPushButton(self.layoutWidget)
        self.pushButton_row_up.setObjectName(u"pushButton_row_up")
        self.pushButton_row_up.setText(u"")
        icon6 = QIcon()
        iconThemeName = u"arrow-up"
        if QIcon.hasThemeIcon(iconThemeName):
            icon6 = QIcon.fromTheme(iconThemeName)
        else:
            icon6.addFile(u".", QSize(), QIcon.Mode.Normal, QIcon.State.Off)

        self.pushButton_row_up.setIcon(icon6)
        self.pushButton_row_up.setFlat(True)

        self.horizontalLayout.addWidget(self.pushButton_row_up)

        self.pushButton_row_down = QPushButton(self.layoutWidget)
        self.pushButton_row_down.setObjectName(u"pushButton_row_down")
        self.pushButton_row_down.setText(u"")
        icon7 = QIcon()
        iconThemeName = u"arrow-down"
        if QIcon.hasThemeIcon(iconThemeName):
            icon7 = QIcon.fromTheme(iconThemeName)
        else:
            icon7.addFile(u".", QSize(), QIcon.Mode.Normal, QIcon.State.Off)

        self.pushButton_row_down.setIcon(icon7)
        self.pushButton_row_down.setFlat(True)

        self.horizontalLayout.addWidget(self.pushButton_row_down)

        self.pushButton_new = QPushButton(self.layoutWidget)
        self.pushButton_new.setObjectName(u"pushButton_new")
        self.pushButton_new.setText(u"<new>")
        self.pushButton_new.setCheckable(True)
        self.pushButton_new.setFlat(True)

        self.horizontalLayout.addWidget(self.pushButton_new)

        self.pushButton_delete = QPushButton(self.layoutWidget)
        self.pushButton_delete.setObjectName(u"pushButton_delete")
        self.pushButton_delete.setText(u"")
        icon8 = QIcon()
        iconThemeName = u"edit-delete"
        if QIcon.hasThemeIcon(iconThemeName):
            icon8 = QIcon.fromTheme(iconThemeName)
        else:
            icon8.addFile(u".", QSize(), QIcon.Mode.Normal, QIcon.State.Off)

        self.pushButton_delete.setIcon(icon8)
        self.pushButton_delete.setFlat(True)

        self.horizontalLayout.addWidget(self.pushButton_delete)

        self.pushButton_undelete = QPushButton(self.layoutWidget)
        self.pushButton_undelete.setObjectName(u"pushButton_undelete")
        self.pushButton_undelete.setText(u"<undelete>")
        self.pushButton_undelete.setFlat(True)

        self.horizontalLayout.addWidget(self.pushButton_undelete)

        self.pushButton_reset = QPushButton(self.layoutWidget)
        self.pushButton_reset.setObjectName(u"pushButton_reset")
        icon9 = QIcon()
        iconThemeName = u"edit-reset"
        if QIcon.hasThemeIcon(iconThemeName):
            icon9 = QIcon.fromTheme(iconThemeName)
        else:
            icon9.addFile(u".", QSize(), QIcon.Mode.Normal, QIcon.State.Off)

        self.pushButton_reset.setIcon(icon9)
        self.pushButton_reset.setFlat(True)

        self.horizontalLayout.addWidget(self.pushButton_reset)

        self.pushButton_reset_all = QPushButton(self.layoutWidget)
        self.pushButton_reset_all.setObjectName(u"pushButton_reset_all")
        icon10 = QIcon()
        iconThemeName = u"document-revert"
        if QIcon.hasThemeIcon(iconThemeName):
            icon10 = QIcon.fromTheme(iconThemeName)
        else:
            icon10.addFile(u".", QSize(), QIcon.Mode.Normal, QIcon.State.Off)

        self.pushButton_reset_all.setIcon(icon10)
        self.pushButton_reset_all.setFlat(True)

        self.horizontalLayout.addWidget(self.pushButton_reset_all)

        self.line_2 = QFrame(self.layoutWidget)
        self.line_2.setObjectName(u"line_2")
        self.line_2.setFrameShape(QFrame.Shape.VLine)
        self.line_2.setFrameShadow(QFrame.Shadow.Sunken)

        self.horizontalLayout.addWidget(self.line_2)

        self.label_2 = QLabel(self.layoutWidget)
        self.label_2.setObjectName(u"label_2")

        self.horizontalLayout.addWidget(self.label_2)

        self.comboBox_ocr_language = CComboBox(self.layoutWidget)
        self.comboBox_ocr_language.setObjectName(u"comboBox_ocr_language")

        self.horizontalLayout.addWidget(self.comboBox_ocr_language)

        self.pushButton_done = QPushButton(self.layoutWidget)
        self.pushButton_done.setObjectName(u"pushButton_done")

        self.horizontalLayout.addWidget(self.pushButton_done)

        self.horizontalLayout.setStretch(0, 1)

        self.verticalLayout.addLayout(self.horizontalLayout)

        self.splitter_side = QSplitter(self.layoutWidget)
        self.splitter_side.setObjectName(u"splitter_side")
        self.splitter_side.setOrientation(Qt.Horizontal)
        self.image_viewer = BubbleImageViewer(self.splitter_side)
        self.image_viewer.setObjectName(u"image_viewer")
        self.splitter_side.addWidget(self.image_viewer)
        self.tableWidget_ocr = CTableWidget(self.splitter_side)
        if (self.tableWidget_ocr.columnCount() < 2):
            self.tableWidget_ocr.setColumnCount(2)
        __qtablewidgetitem = QTableWidgetItem()
        self.tableWidget_ocr.setHorizontalHeaderItem(0, __qtablewidgetitem)
        __qtablewidgetitem1 = QTableWidgetItem()
        self.tableWidget_ocr.setHorizontalHeaderItem(1, __qtablewidgetitem1)
        self.tableWidget_ocr.setObjectName(u"tableWidget_ocr")
        self.tableWidget_ocr.setAcceptDrops(True)
        self.tableWidget_ocr.setEditTriggers(QAbstractItemView.AllEditTriggers)
        self.tableWidget_ocr.setTabKeyNavigation(False)
        self.tableWidget_ocr.setProperty("showDropIndicator", True)
        self.tableWidget_ocr.setDragEnabled(True)
        self.tableWidget_ocr.setDragDropOverwriteMode(False)
        self.tableWidget_ocr.setDragDropMode(QAbstractItemView.NoDragDrop)
        self.tableWidget_ocr.setDefaultDropAction(Qt.MoveAction)
        self.tableWidget_ocr.setAlternatingRowColors(True)
        self.tableWidget_ocr.setSelectionMode(QAbstractItemView.SingleSelection)
        self.tableWidget_ocr.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.splitter_side.addWidget(self.tableWidget_ocr)
        self.tableWidget_ocr.horizontalHeader().setDefaultSectionSize(80)
        self.tableWidget_ocr.horizontalHeader().setStretchLastSection(True)
        self.tableWidget_ocr.verticalHeader().setVisible(False)

        self.verticalLayout.addWidget(self.splitter_side)

        self.verticalLayout.setStretch(1, 1)
        self.splitter.addWidget(self.layoutWidget)

        self.verticalLayout_2.addWidget(self.splitter)


        self.retranslateUi(OcrReview)

        QMetaObject.connectSlotsByName(OcrReview)
    # setupUi

    def retranslateUi(self, OcrReview):
        self.label.setText(QCoreApplication.translate("OcrReview", u"Icon Size:", None))
#if QT_CONFIG(tooltip)
        self.pushButton_prev.setToolTip(QCoreApplication.translate("OcrReview", u"Previous image", None))
#endif // QT_CONFIG(tooltip)
#if QT_CONFIG(tooltip)
        self.pushButton_next.setToolTip(QCoreApplication.translate("OcrReview", u"Next image", None))
#endif // QT_CONFIG(tooltip)
#if QT_CONFIG(tooltip)
        self.pushButton_zoom_in.setToolTip(QCoreApplication.translate("OcrReview", u"Zoom in", None))
#endif // QT_CONFIG(tooltip)
        self.pushButton_zoom_in.setText("")
#if QT_CONFIG(tooltip)
        self.pushButton_zoom_out.setToolTip(QCoreApplication.translate("OcrReview", u"Zoom out", None))
#endif // QT_CONFIG(tooltip)
        self.pushButton_zoom_out.setText("")
#if QT_CONFIG(tooltip)
        self.pushButton_zoom_reset.setToolTip(QCoreApplication.translate("OcrReview", u"Reset zoom", None))
#endif // QT_CONFIG(tooltip)
        self.pushButton_zoom_reset.setText("")
#if QT_CONFIG(tooltip)
        self.pushButton_zoom_fit.setToolTip(QCoreApplication.translate("OcrReview", u"Zoom to fit", None))
#endif // QT_CONFIG(tooltip)
        self.pushButton_zoom_fit.setText("")
        self.comboBox_view_mode.setItemText(0, QCoreApplication.translate("OcrReview", u"With Boxes", None))
        self.comboBox_view_mode.setItemText(1, QCoreApplication.translate("OcrReview", u"Original", None))

#if QT_CONFIG(tooltip)
        self.pushButton_row_up.setToolTip(QCoreApplication.translate("OcrReview", u"Move box order up", None))
#endif // QT_CONFIG(tooltip)
#if QT_CONFIG(tooltip)
        self.pushButton_row_down.setToolTip(QCoreApplication.translate("OcrReview", u"Move box order down", None))
#endif // QT_CONFIG(tooltip)
#if QT_CONFIG(tooltip)
        self.pushButton_new.setToolTip(QCoreApplication.translate("OcrReview", u"Add a new box", None))
#endif // QT_CONFIG(tooltip)
#if QT_CONFIG(tooltip)
        self.pushButton_delete.setToolTip(QCoreApplication.translate("OcrReview", u"Delete current box", None))
#endif // QT_CONFIG(tooltip)
#if QT_CONFIG(tooltip)
        self.pushButton_undelete.setToolTip(QCoreApplication.translate("OcrReview", u"Recover this deleted box", None))
#endif // QT_CONFIG(tooltip)
#if QT_CONFIG(tooltip)
        self.pushButton_reset.setToolTip(QCoreApplication.translate("OcrReview", u"Reset changes to this image", None))
#endif // QT_CONFIG(tooltip)
        self.pushButton_reset.setText("")
        self.pushButton_reset_all.setText(QCoreApplication.translate("OcrReview", u"Reset All", None))
        self.label_2.setText(QCoreApplication.translate("OcrReview", u"OCR new boxes:", None))
        self.pushButton_done.setText(QCoreApplication.translate("OcrReview", u"Done", None))
        ___qtablewidgetitem = self.tableWidget_ocr.horizontalHeaderItem(0)
        ___qtablewidgetitem.setText(QCoreApplication.translate("OcrReview", u"Box", None));
        ___qtablewidgetitem1 = self.tableWidget_ocr.horizontalHeaderItem(1)
        ___qtablewidgetitem1.setText(QCoreApplication.translate("OcrReview", u"Text", None));
        pass
    # retranslateUi




import os
from copy import deepcopy, copy
from functools import partial
from multiprocessing import Pool
from pathlib import Path
from typing import Callable

import torch
from loguru import logger

import pcleaner.config as cfg
import pcleaner.denoiser as dn
import pcleaner.gui.ctd_interface_gui as ctm
import pcleaner.gui.gui_utils as gu
import pcleaner.gui.image_file as imf
import pcleaner.gui.worker_thread as wt
import pcleaner.image_export as ie
import pcleaner.inpainting as ip
import pcleaner.masker as ma
import pcleaner.ocr.ocr as ocr
import pcleaner.output_structures as ost
import pcleaner.preprocessor as pp
import pcleaner.structures as st
from pcleaner import model_downloader as md
from pcleaner.config import LayeredExport
from pcleaner.helpers import tr


def generate_output(
    image_objects: list[imf.ImageFile],
    target_outputs: list[ost.Output],
    output_dir: Path | None,
    config: cfg.Config,
    ocr_processor: ocr.OCREngineFactory | None,
    progress_callback: ost.ProgressSignal,
    abort_flag: wt.SharableFlag,
    debug: bool = False,
):
    """
    Process all the images in the given image objects list until the target step is reached.
    The paths inside the image objects are updated as the processing progresses.
    If the profile checksum of a step hasn't changed, the step is skipped.

    In GUI mode, the masker and denoiser never write the output directly, but instead
    place them in the cache directory. From there they may be copied to the output directory,
    to ensure that effort isn't duplicated where possible. For example, if the output was already
    generated for the viewer, it doesn't need to be generated again for the final output, or when
    only changing the output path for the final output.

    The ocr model (or None) is passed in as arguments, so it can be reused
    across multiple runs. This is done because in gui mode the model does not need to be loaded
    repeatedly for single image runs, making iterations much faster, due to the slow
    initializations of the model taking several seconds.

    When setting the output dir to None, intermediate steps are saved to disk, offering
    additional visualizations that aren't necessary when you aren't viewing image details.

    :param image_objects: The image objects to process. These contain the image paths.
    :param target_outputs: The target outputs to reach, eg. the clean mask and image.
        Multiple outputs are only supported for those within the same step, as it's meant to accommodate
        the full run parameters.
    :param output_dir: The directory to save the outputs to. If None, the output remains in the cache directory.
    :param config: The config to use.
    :param ocr_processor: The ocr models to use.
    :param progress_callback: A callback to report progress to the gui.
    :param abort_flag: A flag that is set to True when the thread should abort.
    :param debug: If True, debug messages are printed.
    """

    # We will check before each step and during each iteration of the step to see if the thread should abort.
    # The check is NOT done right before checking the outputs to avoid throwing away perfectly fine work.
    # This means that the abort check shouldn't interrupt the processing step on an image, leaving it in a
    # dirty state.
    def check_abortion() -> None:
        """
        Check if the thread should abort.
        If so, signal abortion to the progress callback and raise an AbortException.
        """
        if abort_flag.get():
            progress_callback.emit(
                ost.ProgressData(
                    0,
                    [],
                    ost.Step.output,
                    ost.ProgressType.aborted,
                )
            )
            raise wt.Abort()

    progress_callback.emit(
        ost.ProgressData(
            0,
            [],
            ost.Step.text_detection,
            ost.ProgressType.start,
        )
    )

    # Create an independent copy of the profile, since this instance is shared and mutated across threads.
    # Due to the checksum only being generated at the end, we don't want to erroneously claim this output
    # matches a profile that was changed in the meantime.
    profile = deepcopy(config.current_profile)

    cache_dir = config.get_cleaner_cache_dir()

    # Pick out the highest output we need to reach.
    # Other outputs are only considered if they are in the same step, used for final output runs.
    target_output = max(target_outputs)

    # Get the text detector model path from the config.
    cuda = torch.cuda.is_available()
    text_detector_model_path = config.get_model_path(cuda)

    def step_needs_to_be_rerun_closure(current_step: ost.Step) -> Callable[[imf.ImageFile], bool]:
        """
        Check if any of the outputs for the given image object need to be rerun.

        :return: A function that takes an image object and returns True if the step needs to be rerun for it.
        """

        def step_changed(image_object: imf.ImageFile) -> bool:
            nonlocal profile
            nonlocal target_outputs
            nonlocal current_step
            # holy shit this is cursed I need to think about it more.
            # I think I got it sorted now...
            # Pretty sure it works :)
            # It didn't.
            # Now it is fr fr

            target_outputs_in_the_current_step: tuple[ost.Output] = tuple(
                filter(lambda o: o in ost.step_to_output[current_step], reversed(target_outputs))
            )

            if target_outputs_in_the_current_step:
                # We need to check each output we care about to know if we gotta rerun the step.
                return any(
                    image_object.outputs[o].is_changed(profile)
                    for o in target_outputs_in_the_current_step
                )
            else:
                # In this case we just need to know if the step's representative is complete.
                representative: ost.Output = ost.get_output_representing_step(current_step)
                return image_object.outputs[representative].is_changed(profile)

        return step_changed

    def update_output(image_object: imf.ImageFile, current_output: ost.Output) -> None:
        """
        Update the output of the given image object.
        Check if the file actually exists, and if it does, update the output path.

        :param image_object: The image object to update.
        :param current_output: The output to update.
        """
        nonlocal profile, cache_dir

        path_gen = ost.OutputPathGenerator(image_object.path, cache_dir, image_object.uuid)
        path = path_gen.for_output(current_output)

        if path.is_file():
            image_object.outputs[current_output].update(path, profile)

    # ============================================== Text Detection ==============================================

    check_abortion()

    step_text_detector_images = tuple(
        filter(step_needs_to_be_rerun_closure(ost.Step.text_detection), image_objects)
    )

    if step_text_detector_images:
        logger.info(
            f"Running text detection AI model for {len(step_text_detector_images)} images..."
        )

        visualize_raw_boxes = ost.Output.raw_boxes in target_outputs

        try:
            ctm.model2annotations_gui(
                profile.general,
                profile.text_detector,
                text_detector_model_path,
                step_text_detector_images,
                cache_dir,
                no_text_detection=target_output == ost.Output.input,
                visualize_raw_boxes=visualize_raw_boxes,
                partial_progress_data=partial(
                    ost.ProgressData,
                    len(step_text_detector_images),
                    target_outputs,
                    ost.Step.text_detection,
                ),
                progress_callback=progress_callback if target_output != ost.Output.input else None,
                abort_flag=abort_flag,
            )
        except wt.Abort:
            # Send the progress callback a signal that the process was aborted,
            # and then raise the exception again to stop the processing.
            check_abortion()

        for image_obj in step_text_detector_images:
            for output in ost.step_to_output[ost.Step.text_detection]:
                update_output(image_obj, output)

        progress_callback.emit(
            ost.ProgressData(
                len(step_text_detector_images),
                target_outputs,
                ost.Step.text_detection,
                ost.ProgressType.textDetection_done,
            )
        )

    # ============================================== Preprocessing ==============================================

    check_abortion()

    if target_output > ost.get_output_representing_step(ost.Step.text_detection):
        step_preprocessor_images = tuple(
            filter(step_needs_to_be_rerun_closure(ost.Step.preprocessor), image_objects)
        )

        if step_preprocessor_images:
            logger.info(f"Running preprocessing for {len(step_preprocessor_images)} images...")

            progress_callback.emit(
                ost.ProgressData(
                    len(step_preprocessor_images),
                    target_outputs,
                    ost.Step.preprocessor,
                    ost.ProgressType.begin_step,
                )
            )

            ocr_analytics: list[st.OCRAnalytic] = []
            # Find all the json files associated with the images.
            for image_obj in step_preprocessor_images:
                check_abortion()
                path_gen = ost.OutputPathGenerator(image_obj.path, cache_dir, image_obj.uuid)
                ocr_analytic = pp.prep_json_file(
                    path_gen.raw_json,
                    preprocessor_conf=profile.preprocessor,
                    cache_masks=target_output in ost.step_to_output[ost.Step.preprocessor]
                    or output_dir is None,
                    ocr_engine_factory=ocr_processor if profile.preprocessor.ocr_enabled else None,
                )

                if ocr_analytic is not None:
                    ocr_analytics.append(ocr_analytic)

                progress_callback.emit(
                    ost.ProgressData(
                        len(step_preprocessor_images),
                        target_outputs,
                        ost.Step.preprocessor,
                        ost.ProgressType.incremental,
                    )
                )

            # Update the outputs of the image objects, skipping Output.ocr.
            for image_obj in step_preprocessor_images:
                update_output(image_obj, ost.Output.initial_boxes)
                update_output(image_obj, ost.Output.final_boxes)
                update_output(image_obj, ost.Output.clean_json)

            progress_callback.emit(
                ost.ProgressData(
                    len(step_preprocessor_images),
                    target_outputs,
                    ost.Step.preprocessor,
                    ost.ProgressType.analyticsOCR,
                    (ocr_analytics, profile.preprocessor.ocr_max_size),
                )
            )

    # ============================================== Masker ==============================================

    check_abortion()

    if target_output > ost.get_output_representing_step(ost.Step.preprocessor):
        step_masker_images = tuple(
            filter(step_needs_to_be_rerun_closure(ost.Step.masker), image_objects)
        )

        if step_masker_images:
            logger.info(f"Running masker for {len(step_masker_images)} images...")

            progress_callback.emit(
                ost.ProgressData(
                    len(step_masker_images),
                    target_outputs,
                    ost.Step.masker,
                    ost.ProgressType.begin_step,
                )
            )

            # Find all the json files associated with the images.
            json_files = []
            for image_obj in step_masker_images:
                path_gen = ost.OutputPathGenerator(image_obj.path, cache_dir, image_obj.uuid)
                json_files.append(path_gen.clean_json)

            # Pack all the arguments into a dataclass.
            outputs_that_need_masks = (
                ost.Output.box_mask,
                ost.Output.cut_mask,
                ost.Output.mask_layers,
                ost.Output.mask_overlay,
            )
            # If any of out targeted outputs need to have intermediate masks generated,
            # we need to show the masks. Or if we explicitly want to show the masks.
            need_to_show_masks = output_dir is None or any(
                output in outputs_that_need_masks for output in target_outputs
            )

            data = [
                st.MaskerData(
                    json_file,
                    cache_dir,
                    profile.masker,
                    extract_text=ost.Output.isolated_text in target_outputs,
                    show_masks=need_to_show_masks,
                    debug=debug,
                )
                for json_file in json_files
            ]

            masker_analytics_raw = []

            # Check the size of the required pool, run sequentially if the size is 1.
            core_limit = profile.masker.max_threads
            if core_limit == 0:
                core_limit = os.cpu_count()
            pool_size = min(core_limit, len(data))

            if pool_size > 1:
                with Pool(processes=pool_size) as pool:
                    for analytic in pool.imap(ma.mask_page, data):
                        check_abortion()
                        masker_analytics_raw.extend(analytic)

                        progress_callback.emit(
                            ost.ProgressData(
                                len(step_masker_images),
                                target_outputs,
                                ost.Step.masker,
                                ost.ProgressType.incremental,
                            )
                        )
            else:
                for data_obj in data:
                    check_abortion()
                    analytic = ma.mask_page(data_obj)
                    masker_analytics_raw.extend(analytic)

                    progress_callback.emit(
                        ost.ProgressData(
                            len(step_masker_images),
                            target_outputs,
                            ost.Step.masker,
                            ost.ProgressType.incremental,
                        )
                    )

            for image_obj in step_masker_images:
                for output in ost.step_to_output[ost.Step.masker]:
                    update_output(image_obj, output)

            progress_callback.emit(
                ost.ProgressData(
                    len(step_masker_images),
                    target_outputs,
                    ost.Step.masker,
                    ost.ProgressType.analyticsMasker,
                    (masker_analytics_raw, copy(profile.masker)),  # Make a copy to avoid a mutex.
                )
            )

    # ============================================== Denoiser ==============================================

    check_abortion()

    if (
        target_output > ost.get_output_representing_step(ost.Step.masker)
        and profile.denoiser.denoising_enabled
    ):
        step_denoiser_images = tuple(
            filter(step_needs_to_be_rerun_closure(ost.Step.denoiser), image_objects)
        )

        if step_denoiser_images:
            logger.info(f"Running denoiser for {len(step_denoiser_images)} images...")

            progress_callback.emit(
                ost.ProgressData(
                    len(step_denoiser_images),
                    target_outputs,
                    ost.Step.denoiser,
                    ost.ProgressType.begin_step,
                )
            )

            # Find all the json files associated with the images.
            json_files = []
            for image_obj in step_denoiser_images:
                path_gen = ost.OutputPathGenerator(image_obj.path, cache_dir, image_obj.uuid)
                json_files.append(path_gen.mask_data_json)

            # Pack all the arguments into a dataclass.
            data = [
                st.DenoiserData(
                    json_file,
                    cache_dir,
                    profile.denoiser,
                    debug,
                )
                for json_file in json_files
            ]

            denoise_analytics_raw = []

            # Check the size of the required pool, run sequentially if the size is 1.
            core_limit = profile.denoiser.max_threads
            if core_limit == 0:
                core_limit = os.cpu_count()
            pool_size = min(core_limit, len(data))

            if pool_size > 1:
                with Pool() as pool:
                    for analytic in pool.imap(dn.denoise_page, data):
                        check_abortion()
                        denoise_analytics_raw.append(analytic)

                        progress_callback.emit(
                            ost.ProgressData(
                                len(step_denoiser_images),
                                target_outputs,
                                ost.Step.denoiser,
                                ost.ProgressType.incremental,
                            )
                        )
            else:
                for data_obj in data:
                    check_abortion()
                    analytic = dn.denoise_page(data_obj)
                    denoise_analytics_raw.append(analytic)

                    progress_callback.emit(
                        ost.ProgressData(
                            len(step_denoiser_images),
                            target_outputs,
                            ost.Step.denoiser,
                            ost.ProgressType.incremental,
                        )
                    )

            for image_obj in step_denoiser_images:
                for output in ost.step_to_output[ost.Step.denoiser]:
                    update_output(image_obj, output)

            progress_callback.emit(
                ost.ProgressData(
                    len(step_denoiser_images),
                    target_outputs,
                    ost.Step.denoiser,
                    ost.ProgressType.analyticsDenoiser,
                    (
                        denoise_analytics_raw,
                        profile.denoiser.noise_min_standard_deviation,
                        profile.masker.mask_max_standard_deviation,
                    ),
                )
            )

    # ============================================== Inpainting ==============================================

    check_abortion()

    if (
        target_output > ost.get_output_representing_step(ost.Step.denoiser)
        and profile.inpainter.inpainting_enabled
    ):
        step_inpainting_images = tuple(
            filter(step_needs_to_be_rerun_closure(ost.Step.inpainter), image_objects)
        )

        if step_inpainting_images:
            logger.info(f"Running inpainting for {len(step_inpainting_images)} images...")

            progress_callback.emit(
                ost.ProgressData(
                    len(step_inpainting_images),
                    target_outputs,
                    ost.Step.inpainter,
                    ost.ProgressType.begin_step,
                )
            )

            # Find all the json files associated with the images.
            json_files = []
            for image_obj in step_inpainting_images:
                path_gen = ost.OutputPathGenerator(image_obj.path, cache_dir, image_obj.uuid)
                json_files.append((path_gen.clean_json, path_gen.mask_data_json))

            # Check the inpainting model is available.
            if not md.is_inpainting_downloaded(config):
                raise FileNotFoundError(tr("Inpainting model not found."))

            inpainter_model = ip.InpaintingModel(config)

            # Pack all the arguments into a dataclass.
            data = [
                st.InpainterData(
                    page_json_file,
                    mask_json_file,
                    cache_dir,
                    profile.masker,
                    profile.denoiser,
                    profile.inpainter,
                    debug,
                )
                for page_json_file, mask_json_file in json_files
            ]

            # Single threaded due to model loading overhead.
            inpaint_analytics_raw = []
            for data_obj in data:
                check_abortion()
                analytic = ip.inpaint_page(data_obj, inpainter_model)
                inpaint_analytics_raw.append(analytic)

                progress_callback.emit(
                    ost.ProgressData(
                        len(step_inpainting_images),
                        target_outputs,
                        ost.Step.inpainter,
                        ost.ProgressType.incremental,
                    )
                )

            for image_obj in step_inpainting_images:
                for output in ost.step_to_output[ost.Step.inpainter]:
                    update_output(image_obj, output)

            progress_callback.emit(
                ost.ProgressData(
                    len(step_inpainting_images),
                    target_outputs,
                    ost.Step.inpainter,
                    ost.ProgressType.analyticsInpainter,
                    (
                        inpaint_analytics_raw,
                        profile.inpainter.min_inpainting_radius,
                        profile.inpainter.max_inpainting_radius,
                    ),
                )
            )

    # ============================================== Final Output ==============================================

    logger.info(f"Finished processing {len(image_objects)} images.")

    check_abortion()

    if target_output == ost.Output.write_output:
        progress_callback.emit(
            ost.ProgressData(
                len(image_objects),
                target_outputs,
                ost.Step.output,
                ost.ProgressType.begin_step,
            )
        )

        # Pack all the arguments in batches.
        data = [
            (
                image_obj.path,
                cache_dir,
                image_obj.uuid,
                target_outputs,
                output_dir,
                profile.general.preferred_file_type,
                profile.general.preferred_mask_file_type,
                profile.denoiser.denoising_enabled,
                profile.general.layered_export,
            )
            for image_obj in image_objects
        ]

        # Check the size of the required pool, run sequentially if the size is 1.
        core_limit = profile.general.max_threads_export
        if core_limit == 0:
            core_limit = os.cpu_count()
        pool_size = min(core_limit, len(data))

        if pool_size > 1:
            with Pool(processes=pool_size) as pool:
                for _ in pool.imap_unordered(ie.copy_to_output_batched, data):
                    check_abortion()

                    progress_callback.emit(
                        ost.ProgressData(
                            len(image_objects),
                            target_outputs,
                            ost.Step.output,
                            ost.ProgressType.incremental,
                        )
                    )
        else:
            for args in data:
                check_abortion()
                ie.copy_to_output(*args)

                progress_callback.emit(
                    ost.ProgressData(
                        len(image_objects),
                        target_outputs,
                        ost.Step.output,
                        ost.ProgressType.incremental,
                    )
                )

        if profile.general.layered_export == LayeredExport.PSD_BULK:
            ie.bundle_psd(
                output_dir,
                cache_dir,
                [image_object.path for image_object in image_objects],
                [image_object.uuid for image_object in image_objects],
            )

    progress_callback.emit(
        ost.ProgressData(
            0,
            [],
            ost.Step.output,
            ost.ProgressType.end,
        )
    )


def perform_ocr(
    image_objects: list[imf.ImageFile],
    output_file: Path | None,
    csv_output: bool,
    config: cfg.Config,
    ocr_engine_factory: ocr.OCREngineFactory | None,
    progress_callback: ost.ProgressSignal,
    abort_flag: wt.SharableFlag,
    debug: bool = False,
):
    """
    Perform OCR on all the images in the given image objects list.
    If a path is given, the output is written to that path.
    Output is either a plain text file or csv.

    The ocr model (or None) is passed in as arguments, so it can be reused
    across multiple runs. This is done because in gui mode the model does not need to be loaded
    repeatedly for single image runs, making iterations much faster, due to the slow
    initializations of the model taking several seconds.

    When setting the output dir to None, intermediate steps are saved to disk, offering
    additional visualizations that aren't necessary when you aren't viewing image details.

    :param image_objects: The image objects to process. These contain the image paths.
    :param output_file: The directory to save the outputs to. If None, nothing is written.
    :param csv_output: If True, the output is written as a csv file.
    :param config: The config to use.
    :param ocr_engine_factory: The ocr processors to use.
    :param progress_callback: A callback to report progress to the gui.
    :param abort_flag: A flag that is set to True when the thread should abort.
    :param debug: If True, debug messages are printed.
    """

    def check_abortion() -> None:
        """
        Check if the thread should abort.
        If so, signal abortion to the progress callback and raise an AbortException.
        """
        if abort_flag.get():
            progress_callback.emit(
                ost.ProgressData(
                    0,
                    [],
                    ost.Step.output,
                    ost.ProgressType.aborted,
                )
            )
            raise wt.Abort()

    target_outputs = [ost.Output.ocr]

    progress_callback.emit(
        ost.ProgressData(
            0,
            [],
            ost.Step.text_detection,
            ost.ProgressType.start,
        )
    )

    # Create an independent copy of the profile, since this instance is shared and mutated across threads.
    # Due to the checksum only being generated at the end, we don't want to erroneously claim this output
    # matches a profile that was changed in the meantime.
    # Also, we will be editing this profile to get OCR working.
    profile = deepcopy(config.current_profile)

    # Make sure OCR is enabled.
    profile.preprocessor.ocr_enabled = True
    # Make sure the max size is infinite, so no boxes are skipped in the OCR process.
    profile.preprocessor.ocr_max_size = 10**10
    # Set the OCR blacklist pattern to match everything, so all text gets reported in the analytics.
    profile.preprocessor.ocr_blacklist_pattern = ".*"

    cache_dir = config.get_cleaner_cache_dir()

    # Get the text detector model path from the config.
    cuda = torch.cuda.is_available()
    text_detector_model_path = config.get_model_path(cuda)

    def step_needs_to_be_rerun_closure(current_step: ost.Step) -> Callable[[imf.ImageFile], bool]:
        """
        Check if any of the outputs for the given image object need to be rerun.

        :return: A function that takes an image object and returns True if the step needs to be rerun for it.
        """

        def step_changed(image_object: imf.ImageFile) -> bool:
            nonlocal profile
            nonlocal target_outputs
            nonlocal current_step

            target_outputs_in_the_current_step: tuple[ost.Output] = tuple(
                filter(lambda o: o in ost.step_to_output[current_step], reversed(target_outputs))
            )

            if target_outputs_in_the_current_step:
                # We need to check each output we care about to know if we gotta rerun the step.
                return any(
                    image_object.outputs[o].is_changed(profile)
                    for o in target_outputs_in_the_current_step
                )
            else:
                # In this case we just need to know if the step's representative is complete.
                representative: ost.Output = ost.get_output_representing_step(current_step)
                return image_object.outputs[representative].is_changed(profile)

        return step_changed

    def update_output(image_object: imf.ImageFile, current_output: ost.Output) -> None:
        """
        Update the output of the given image object.
        Check if the file actually exists, and if it does, update the output path.

        :param image_object: The image object to update.
        :param current_output: The output to update.
        """
        nonlocal profile, cache_dir

        _path_gen = ost.OutputPathGenerator(image_object.path, cache_dir, image_object.uuid)
        path = _path_gen.for_output(current_output)

        if path.is_file():
            image_object.outputs[current_output].update(path, profile)

    # ============================================== Text Detection ==============================================

    check_abortion()

    step_text_detector_images = tuple(
        filter(step_needs_to_be_rerun_closure(ost.Step.text_detection), image_objects)
    )

    if step_text_detector_images:
        logger.info(
            f"Running text detection AI model for {len(step_text_detector_images)} images..."
        )
        try:
            ctm.model2annotations_gui(
                profile.general,
                profile.text_detector,
                text_detector_model_path,
                step_text_detector_images,
                cache_dir,
                no_text_detection=False,
                visualize_raw_boxes=True,
                partial_progress_data=partial(
                    ost.ProgressData,
                    len(step_text_detector_images),
                    target_outputs,
                    ost.Step.text_detection,
                ),
                progress_callback=progress_callback,
                abort_flag=abort_flag,
            )
        except wt.Abort:
            # Send the progress callback a signal that the process was aborted,
            # and then raise the exception again to stop the processing.
            check_abortion()

        # Update the outputs of the image objects.
        for image_obj in step_text_detector_images:
            for output in ost.step_to_output[ost.Step.text_detection]:
                update_output(image_obj, output)

    # ============================================== Preprocessing ==============================================

    check_abortion()

    logger.info(f"Running preprocessing for {len(image_objects)} images...")

    progress_callback.emit(
        ost.ProgressData(
            len(image_objects),
            target_outputs,
            ost.Step.preprocessor,
            ost.ProgressType.begin_step,
        )
    )

    ocr_analytics: list[st.OCRAnalytic] = []
    # Find all the json files associated with the images.
    for image_obj in image_objects:
        check_abortion()
        path_gen = ost.OutputPathGenerator(image_obj.path, cache_dir, image_obj.uuid)
        ocr_analytic = pp.prep_json_file(
            path_gen.raw_json,
            preprocessor_conf=profile.preprocessor,
            cache_masks=False,
            ocr_engine_factory=ocr_engine_factory,
            cache_masks_ocr=True,
            performing_ocr=True,
        )

        if ocr_analytic is not None:
            ocr_analytics.append(ocr_analytic)

        progress_callback.emit(
            ost.ProgressData(
                len(image_objects),
                target_outputs,
                ost.Step.preprocessor,
                ost.ProgressType.incremental,
            )
        )

    # Update only the raw boxes, the rest are tainted by the forced profile changes.
    for image_obj in image_objects:
        update_output(image_obj, ost.Output.initial_boxes)

    logger.info(f"Finished processing {len(image_objects)} images.")

    # ============================================== Final Output ==============================================

    check_abortion()

    # Output the OCRed text from the analytics.
    text_out = ocr.format_output(
        ocr_analytics,
        csv_output,
        (tr("filename"), tr("startx"), tr("starty"), tr("endx"), tr("endy"), tr("text")),
    )

    text_out = text_out.strip("\n \t")

    if output_file is not None:
        try:
            output_file.parent.mkdir(parents=True, exist_ok=True)
            output_file.write_text(text_out, encoding="utf-8")
            text_out += "\n\n" + tr("Saved detected text to {output_file}").format(
                output_file=output_file
            )
        except OSError:
            text_out += "\n\n" + tr("Failed to write detected text to {output_file}").format(
                output_file=output_file
            )
            gu.show_exception(None, tr("Save Failed"), tr("Failed to write detected text to file."))

    progress_callback.emit(
        ost.ProgressData(
            0,
            [],
            ost.Step.preprocessor,
            ost.ProgressType.outputOCR,
            (text_out, ocr_analytics),
        )
    )

    progress_callback.emit(
        ost.ProgressData(
            0,
            [],
            ost.Step.output,
            ost.ProgressType.end,
        )
    )



from pathlib import Path

from loguru import logger
from manga_ocr import MangaOcr as MangaOcrModel
from PIL import Image

import pcleaner.ocr.supported_languages as osl


class MangaOcr:
    _instance = None
    _model = None
    _init_args = ((), {})

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            logger.info("Creating the MangaOcr instance")
            cls._instance = super(MangaOcr, cls).__new__(cls)
            cls._instance._model = None
            cls._instance._init_args = (args, kwargs)  # store just in case
        return cls._instance

    def __init__(self, *args, **kwargs):
        # Initialization logic is deferred to a separate method
        pass

    @staticmethod
    def langs() -> set[osl.LanguageCode]:
        return {osl.LanguageCode.jpn}

    def initialize_model(self, *args, **kwargs):
        if self._model is None:
            if args or kwargs:
                self._model = MangaOcrModel(*args, **kwargs)
            else:
                init_args = self._init_args
                self._model = MangaOcrModel(*init_args[0], **init_args[1])
        return self._model

    def __call__(self, img_or_path: Image.Image | Path | str, **kwargs) -> str:
        model = self.initialize_model()
        return model(img_or_path)



from pathlib import Path

from PIL import Image
from attrs import frozen
from loguru import logger
from psd_tools import PSDImage
from psd_tools.api.layers import Group, PixelLayer
from psd_tools.constants import Compression

import pcleaner.output_structures as ost
import pcleaner.structures as st
from pcleaner.helpers import tr
from pcleaner.config import LayeredExport


# SUPPORTED_IMG_TYPES = [".jpeg", ".jpg", ".png", ".bmp", ".tiff", ".tif", ".jp2", ".dib", ".webp", ".ppm"]

# Map the file extension to the type constant used in Pillow.
suffix_to_format: dict[str, str] = {
    ".jpg": "JPEG",
    ".jpeg": "JPEG",
    ".png": "PNG",
    ".webp": "WEBP",
    ".tiff": "TIFF",
    ".tif": "TIFF",
    ".bmp": "BMP",
    ".dib": "BMP",
    ".jp2": "JPEG2000",
    ".ppm": "PPM",
}


def save_optimized(
    image: Path | Image.Image, path: Path, original: Path | Image.Image | None = None
) -> None:
    """
    Save the image with optimized settings.

    :param image: The image to save.
    :param path: The path to save the image to, containing a suffix.
    :param original: The original image, used to determine the compression method, if applicable.
    """
    assert image is not None

    # Load the image if it's a path.
    if isinstance(image, Path):
        image = Image.open(image)

    compression_method = None
    original_image_mode = None
    original_dpi = None

    if original is not None:
        if isinstance(original, Path):
            original = Image.open(original)
        if suffix_to_format[path.suffix.lower()] == original.format:
            compression_method = original.info.get("compression", None)

        original_image_mode = original.mode
        original_dpi = original.info.get("dpi", None)

    if original_image_mode is not None:
        image = image.convert(original_image_mode)

    # Image args may be different depending on the image type.
    kwargs = {"optimize": True}

    match path.suffix.lower():
        case ".png":
            kwargs["compress_level"] = 9
        case ".jpg" | ".jpeg":
            kwargs["quality"] = 95
            kwargs["progressive"] = True
        case ".tif" | ".tiff":
            # Use tiff_lzw compression by default, rather than leaving them uncompressed.
            kwargs["compression"] = compression_method if compression_method else "tiff_lzw"

    if original_dpi is not None:
        kwargs["dpi"] = original_dpi
    logger.debug(f"Saving image {path.name} with kwargs: {kwargs}")
    image.save(path, **kwargs)


def copy_to_output_batched(arg_tuple: tuple) -> None:
    copy_to_output(*arg_tuple)


def copy_to_output(
    original_image_path: Path,
    cache_dir: Path,
    uuid_source: str | Path,
    outputs: list[ost.Output],
    output_directory: Path,
    preferred_file_type: str | None,
    preferred_mask_file_type: str,
    denoising_enabled: bool,
    layered_export: LayeredExport,
) -> None:
    """
    Copy or export the outputs from the cache directory to the output directory.
    Output paths and preferred file types are taken into account.

    Supported outputs:
    - Masked image: Output.masked_output
    - Final mask: Output.final_mask
    - Isolated text: Output.isolated_text
    - Denoised image: Output.denoised_output
    - Denoised Mask: Output.denoise_mask
    - Inpainted image: Output.inpainted_output
    - Inpainted Mask: Output.inpainted_mask



    This may raise OSError in various circumstances.

    We need to know if denoising was enabled so that we know to pull in the
    denoised mask when compositing the inpainting mask from the previous mask layers.

    :param original_image_path: The path to the original file to read it's metadata.
    :param cache_dir: Where the cached outputs are.
    :param uuid_source: One of the cached images with the uuid or the uuid itself.
    :param outputs: The outputs to copy.
    :param output_directory: The directory to write the outputs to.
    :param preferred_file_type: Profile setting.
    :param preferred_mask_file_type: Profile setting.
    :param denoising_enabled: Profile setting.
    :param layered_export: Option to create a layered project file from the outputs.
    """

    masker_mask_name = tr("Clean mask", "layered export", "Label for the masking step mask.")
    denoiser_mask_name = tr("Denoised mask", "layered export", "Label for the denoising step mask.")
    inpainter_mask_name = tr(
        "Inpainting mask", "layered export", "Label for the inpainting step mask."
    )

    if output_directory.is_absolute():
        # When absolute, the output directory is used as is.
        final_out_path = output_directory / original_image_path.name
        path_gen = ost.OutputPathGenerator(original_image_path, output_directory, export_mode=True)
    else:
        # Otherwise, the output directory is relative to the original image's parent directory.
        final_out_path = original_image_path.parent / output_directory / original_image_path.name
        path_gen = ost.OutputPathGenerator(
            original_image_path, original_image_path.parent / output_directory, export_mode=True
        )

    final_out_path.parent.mkdir(parents=True, exist_ok=True)

    # Create png paths for the outputs.
    cleaned_out_path = path_gen.clean
    masked_out_path = path_gen.mask
    text_out_path = path_gen.text

    cache_path_gen = ost.OutputPathGenerator(original_image_path, cache_dir, uuid_source)

    # Check what the preferred output format is.
    if preferred_file_type is None:
        # Use the original file type by default.
        cleaned_out_path = cleaned_out_path.with_suffix(original_image_path.suffix)
    else:
        cleaned_out_path = cleaned_out_path.with_suffix(preferred_file_type)

    if preferred_mask_file_type is None:
        # Use png by default.
        masked_out_path = masked_out_path.with_suffix(".png")
        text_out_path = text_out_path.with_suffix(".png")
    else:
        masked_out_path = masked_out_path.with_suffix(preferred_mask_file_type)
        text_out_path = text_out_path.with_suffix(preferred_mask_file_type)

    # Preload the original image.
    original_image = Image.open(original_image_path)
    original_size = original_image.size

    to_layered_images: list[tuple[Image, str]] = []

    def export_single_image(
        image: Path | Image.Image, path: Path, original: Path | Image.Image | None = None
    ) -> None:
        # Pass through to save_optimized if not creating layered project files.
        if layered_export == LayeredExport.NONE:
            save_optimized(image, path, original)

    def add_image_layer(image: Image, name: str) -> None:
        if layered_export != LayeredExport.NONE:
            # Copy the image to prevent mutations.
            to_layered_images.append((image.copy(), name))

    # Output optimized images for all requested outputs.
    if ost.Output.masked_output in outputs:
        export_single_image(
            cache_path_gen.clean,
            cleaned_out_path,
            original_image,
        )

    if ost.Output.final_mask in outputs:
        # First scale the output mask to the original image size.
        final_mask = Image.open(cache_path_gen.combined_mask)
        final_mask = final_mask.resize(original_size, Image.NEAREST)

        add_image_layer(final_mask.convert("RGBA"), masker_mask_name)

        export_single_image(final_mask, masked_out_path)

    if ost.Output.isolated_text in outputs:
        save_optimized(cache_path_gen.text, text_out_path)

    if ost.Output.denoised_output in outputs:
        export_single_image(
            cache_path_gen.clean_denoised,
            cleaned_out_path,
            original_image,
        )

    if ost.Output.denoise_mask in outputs:
        # Special case: Here we need to take the final mask, scale it up, and then paste this on top.
        final_mask = Image.open(cache_path_gen.combined_mask)
        final_mask = final_mask.resize(original_size, Image.BILINEAR)
        denoised_mask = Image.open(cache_path_gen.noise_mask)
        # Ensure both images are RGBA to safely alpha composite them.
        final_mask = final_mask.convert("RGBA")
        denoised_mask = denoised_mask.convert("RGBA")

        add_image_layer(final_mask, masker_mask_name)
        add_image_layer(denoised_mask, denoiser_mask_name)

        final_mask.alpha_composite(denoised_mask)

        export_single_image(final_mask, masked_out_path)

    if ost.Output.inpainted_output in outputs:
        export_single_image(
            cache_path_gen.clean_inpaint,
            cleaned_out_path,
            original_image,
        )

    if ost.Output.inpainted_mask in outputs:
        # Special case: Here we need to take the final mask, scale it up, and then paste the denoising
        final_mask = Image.open(cache_path_gen.combined_mask)
        final_mask = final_mask.resize(original_size, Image.NEAREST)
        final_mask = final_mask.convert("RGBA")

        add_image_layer(final_mask, masker_mask_name)

        if denoising_enabled:
            denoised_mask = Image.open(cache_path_gen.noise_mask)
            denoised_mask = denoised_mask.convert("RGBA")
            final_mask.alpha_composite(denoised_mask)

            add_image_layer(denoised_mask, denoiser_mask_name)

        inpainted_mask = Image.open(cache_path_gen.inpainting)
        inpainted_mask = inpainted_mask.convert("RGBA")

        add_image_layer(inpainted_mask, inpainter_mask_name)

        final_mask.alpha_composite(inpainted_mask)

        export_single_image(final_mask, masked_out_path)

    # Handle special layering cases.
    if layered_export == LayeredExport.PSD_PER_IMAGE:
        export_to_psd(path_gen.psd, original_image.convert("RGBA"), to_layered_images)

    if layered_export == LayeredExport.PSD_BULK:
        export_to_psd(
            ost.OutputPathGenerator(original_image_path, cache_dir, uuid_source).psd,
            original_image.convert("RGBA"),
            to_layered_images,
        )


@frozen
class ExportTarget:
    original_path: Path
    uuid: str
    outputs: list[ost.Output]


def discover_viable_outputs(
    search_dir: Path,
    cleaned_outputs_whitelist: list[ost.Output],
    masked_outputs_whitelist: list[ost.Output],
    text_outputs_whitelist: list[ost.Output],
) -> list[ExportTarget]:
    """
    We need to export everything that is available for export, using the highest stage
    that isn't blocked by the profile and skip flags.
    Precedence: masker < denoiser < inpainter
    Viable candidates need at least a masked version, so the data is gathered from
    all available #clean.json files. Higher stage outputs are probed from there.



    :param search_dir: The cache directory to search in.
    :param cleaned_outputs_whitelist: Check for the highest output among these.
    :param masked_outputs_whitelist: Check for the highest output among these.
    :param text_outputs_whitelist: Check for the highest output among these.
    """
    export_targets = []

    json_files = list(Path(search_dir).glob("*#clean.json"))
    for file in json_files:
        page_data = st.PageData.from_json(file.read_text(encoding="utf-8"))

        cached_image_path = Path(page_data.image_path)
        path_gen = ost.OutputPathGenerator(
            Path(page_data.original_path), cached_image_path.parent, cached_image_path
        )

        # Iterate through all output types in descending order, starting with the highest
        # priority and checking if that file exists. Stop at the first one that exists, if any,
        # then moving on to the next category.
        target_outputs = []
        for out in cleaned_outputs_whitelist[::-1]:
            if path_gen.for_output(out).is_file():
                target_outputs.append(out)
                break
        for out in masked_outputs_whitelist[::-1]:
            if path_gen.for_output(out).is_file():
                target_outputs.append(out)
                break
        for out in text_outputs_whitelist[::-1]:
            if path_gen.for_output(out).is_file():
                target_outputs.append(out)
                break

        if target_outputs:
            logger.debug(
                f"Exporting image {path_gen.original_path.name} for outputs: {target_outputs}"
            )
        else:
            logger.debug(
                f"Skipping export of image {path_gen.original_path.name} due to no valid outputs being generated."
            )
            continue

        export_targets.append(ExportTarget(path_gen.original_path, path_gen.uuid, target_outputs))

    return export_targets


def export_to_psd(path: Path, original_image: Image, masks: list[tuple[Image, str]]) -> None:
    """
    Export the images and masks to a PSD file.

    :param path: The path to write the PSD file to.
    :param original_image: The original image to use as the base layer.
    :param masks: The masks and names to layer on top of the base.
    """

    psd = PSDImage.new("RGBA", (original_image.width, original_image.height))

    base_layer = PixelLayer.frompil(original_image, psd, "Base image", compression=Compression.ZIP)
    psd.append(base_layer)

    group_layer = Group.new("Masks")
    psd.append(group_layer)

    for mask, name in masks:
        layer = PixelLayer.frompil(mask, psd, name, compression=Compression.ZIP)
        group_layer.append(layer)

    psd.save(path)


def bundle_psd(
    output_directory: Path, cache_dir: Path, image_paths: list[Path], uuids: list[str]
) -> None:
    """
    Bundle all the PSDs into a single PSD file.

    :param output_directory: The directory to write the PSD file into.
    :param cache_dir: The directory where the cached PSDs are.
    :param image_paths: The paths to the original images.
    :param uuids: The UUIDs of the images used to generate the cache paths.
    """
    if not image_paths:
        logger.error("No images to bundle PSDs for.")
        return

    if output_directory.is_absolute():
        # When absolute, the output directory is used as is.
        path_gen = ost.OutputPathGenerator(image_paths[0], output_directory, export_mode=True)
    else:
        # Otherwise, the output directory is relative to the original image's parent directory.
        path_gen = ost.OutputPathGenerator(
            image_paths[0], image_paths[0].parent / output_directory, export_mode=True
        )

    bulk_psd_path = path_gen.psd_bulk

    cached_images = []

    for image_path, uuid in zip(image_paths, uuids):
        cached_images.append(ost.OutputPathGenerator(image_path, cache_dir, uuid).psd)

    psds = []
    pages = []

    i = 0
    for cache_image in cached_images:
        psds.append(PSDImage.open(cache_image))
        pages.insert(
            0,
            Group.group_layers(
                psds[i][:],
                tr("Page", "layered export", "Label for Page X in a layered project file.")
                + " {:03d}".format(i + 1),
                open_folder=False,
            ),
        )
        i += 1

    psd_bulk = PSDImage.new("RGBA", psds[0].size, depth=psds[0].depth)

    for page in pages:
        psd_bulk.append(page)

    psd_bulk.save(bulk_psd_path)



import PySide6.QtCore as Qc
import PySide6.QtWidgets as Qw
from loguru import logger


class StateSaver:
    """
    A class to save and restore the state of various widgets
    using the QSettings API.
    """

    widgets: set[Qw.QWidget]
    settings: Qc.QSettings
    name: str  # The namespace in the settings file.
    no_save: bool

    def __init__(self, name: str) -> None:
        self.name = name
        self.widgets = set()
        self.settings = Qc.QSettings("pcleaner", "window_state")
        self.no_save = False

    def register(self, *widget: Qw.QWidget):
        """
        Register a widget to be saved and restored.
        """
        self.widgets.update(widget)

    def save(self):
        """
        Save the state of all registered widgets.
        """
        if self.no_save:
            return

        self.settings.beginGroup(self.name)
        try:
            for widget in self.widgets:
                if isinstance(widget, Qw.QSplitter):
                    self.settings.setValue(widget.objectName(), widget.saveState())
                elif isinstance(widget, Qw.QCheckBox):
                    self.settings.setValue(widget.objectName(), widget.isChecked())
                elif isinstance(widget, Qw.QRadioButton):
                    self.settings.setValue(widget.objectName(), widget.isChecked())
                elif isinstance(widget, Qw.QLineEdit):
                    self.settings.setValue(widget.objectName(), widget.text())
                elif isinstance(widget, Qw.QComboBox):
                    self.settings.setValue(widget.objectName(), widget.currentIndex())
                elif isinstance(widget, Qw.QSpinBox):
                    self.settings.setValue(widget.objectName(), widget.value())
                elif isinstance(widget, Qw.QTableWidget):
                    self.settings.setValue(
                        widget.objectName(), widget.horizontalHeader().saveState()
                    )
                elif isinstance(widget, Qw.QAbstractSlider):
                    self.settings.setValue(widget.objectName(), widget.value())
                elif isinstance(widget, Qw.QWidget):
                    self.settings.setValue(widget.objectName(), widget.saveGeometry())
                else:
                    logger.warning(f"Unknown widget type: {widget}")
        except Exception as e:
            logger.error(f"Failed to save widget state: {e} for {self.name}")
        finally:
            self.settings.endGroup()
        logger.debug(f"Saved state for {self.name}")

    # noinspection PyTypeChecker
    def restore(self):
        """
        Restore the state of all registered widgets.
        """
        self.settings.beginGroup(self.name)
        try:
            for widget in self.widgets:
                # Skip if there is nothing to load.
                if not self.settings.contains(widget.objectName()):
                    continue
                if isinstance(widget, Qw.QSplitter):
                    widget.restoreState(self.settings.value(widget.objectName()))
                elif isinstance(widget, Qw.QCheckBox):
                    value = self.settings.value(widget.objectName())
                    widget.setChecked(value == "true")
                elif isinstance(widget, Qw.QRadioButton):
                    value = self.settings.value(widget.objectName())
                    widget.setChecked(value == "true")
                    if value == "true":
                        widget.clicked.emit()
                elif isinstance(widget, Qw.QLineEdit):
                    widget.setText(self.settings.value(widget.objectName()))
                elif isinstance(widget, Qw.QComboBox):
                    widget.setCurrentIndex(int(self.settings.value(widget.objectName())))
                elif isinstance(widget, Qw.QSpinBox):
                    widget.setValue(int(self.settings.value(widget.objectName())))
                elif isinstance(widget, Qw.QTableWidget):
                    widget.horizontalHeader().restoreState(self.settings.value(widget.objectName()))
                elif isinstance(widget, Qw.QAbstractSlider):
                    widget.setValue(int(self.settings.value(widget.objectName())))
                elif isinstance(widget, Qw.QWidget):
                    widget.restoreGeometry(self.settings.value(widget.objectName()))
                else:
                    logger.warning(f"Unknown widget type: {widget}")
        except Exception as e:
            logger.error(f"Failed to restore widget state: {e} for {self.name}")
        finally:
            self.settings.endGroup()
        logger.debug(f"Restored state for {self.name}")

    def delete_all(self):
        """
        Delete all saved settings.
        """
        self.settings.clear()
        self.no_save = True
        logger.debug(f"Deleted all saved settings for {self.name}")



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'IssueReporter.ui'
##
## Created by: Qt User Interface Compiler version 6.7.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QDialog, QHBoxLayout, QLabel,
    QPushButton, QSizePolicy, QSpacerItem, QVBoxLayout,
    QWidget)

from pcleaner.gui.CustomQ.CComboBox import CComboBox
from pcleaner.gui.log_viewer import LogViewer

class Ui_IssueReporter(object):
    def setupUi(self, IssueReporter):
        if not IssueReporter.objectName():
            IssueReporter.setObjectName(u"IssueReporter")
        IssueReporter.resize(1200, 700)
        self.verticalLayout = QVBoxLayout(IssueReporter)
        self.verticalLayout.setSpacing(6)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.label = QLabel(IssueReporter)
        self.label.setObjectName(u"label")

        self.verticalLayout.addWidget(self.label)

        self.horizontalLayout_2 = QHBoxLayout()
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.label_logs_view = QLabel(IssueReporter)
        self.label_logs_view.setObjectName(u"label_logs_view")

        self.horizontalLayout_2.addWidget(self.label_logs_view)

        self.comboBox_sessions = CComboBox(IssueReporter)
        self.comboBox_sessions.setObjectName(u"comboBox_sessions")

        self.horizontalLayout_2.addWidget(self.comboBox_sessions)


        self.verticalLayout.addLayout(self.horizontalLayout_2)

        self.log_viewer = LogViewer(IssueReporter)
        self.log_viewer.setObjectName(u"log_viewer")
        self.log_viewer.setPlainText(u"")

        self.verticalLayout.addWidget(self.log_viewer)

        self.horizontalLayout_3 = QHBoxLayout()
        self.horizontalLayout_3.setSpacing(12)
        self.horizontalLayout_3.setObjectName(u"horizontalLayout_3")
        self.label_2 = QLabel(IssueReporter)
        self.label_2.setObjectName(u"label_2")
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Maximum, QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_2.sizePolicy().hasHeightForWidth())
        self.label_2.setSizePolicy(sizePolicy)

        self.horizontalLayout_3.addWidget(self.label_2)

        self.label_log_path = QLabel(IssueReporter)
        self.label_log_path.setObjectName(u"label_log_path")
        self.label_log_path.setText(u"<log file path>")
        self.label_log_path.setTextInteractionFlags(Qt.LinksAccessibleByMouse|Qt.TextSelectableByKeyboard|Qt.TextSelectableByMouse)

        self.horizontalLayout_3.addWidget(self.label_log_path)


        self.verticalLayout.addLayout(self.horizontalLayout_3)

        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.label_name_hidden = QLabel(IssueReporter)
        self.label_name_hidden.setObjectName(u"label_name_hidden")
        self.label_name_hidden.setText(u"<name was hidden>")

        self.horizontalLayout.addWidget(self.label_name_hidden)

        self.horizontalSpacer = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout.addItem(self.horizontalSpacer)

        self.pushButton_clipboard = QPushButton(IssueReporter)
        self.pushButton_clipboard.setObjectName(u"pushButton_clipboard")
        icon = QIcon()
        iconThemeName = u"edit-copy"
        if QIcon.hasThemeIcon(iconThemeName):
            icon = QIcon.fromTheme(iconThemeName)
        else:
            icon.addFile(u".", QSize(), QIcon.Mode.Normal, QIcon.State.Off)

        self.pushButton_clipboard.setIcon(icon)

        self.horizontalLayout.addWidget(self.pushButton_clipboard)

        self.pushButton_open_issues = QPushButton(IssueReporter)
        self.pushButton_open_issues.setObjectName(u"pushButton_open_issues")
        icon1 = QIcon()
        iconThemeName = u"internet-services"
        if QIcon.hasThemeIcon(iconThemeName):
            icon1 = QIcon.fromTheme(iconThemeName)
        else:
            icon1.addFile(u".", QSize(), QIcon.Mode.Normal, QIcon.State.Off)

        self.pushButton_open_issues.setIcon(icon1)

        self.horizontalLayout.addWidget(self.pushButton_open_issues)

        self.pushButton_close = QPushButton(IssueReporter)
        self.pushButton_close.setObjectName(u"pushButton_close")
        icon2 = QIcon()
        iconThemeName = u"window-close"
        if QIcon.hasThemeIcon(iconThemeName):
            icon2 = QIcon.fromTheme(iconThemeName)
        else:
            icon2.addFile(u".", QSize(), QIcon.Mode.Normal, QIcon.State.Off)

        self.pushButton_close.setIcon(icon2)

        self.horizontalLayout.addWidget(self.pushButton_close)


        self.verticalLayout.addLayout(self.horizontalLayout)


        self.retranslateUi(IssueReporter)

        QMetaObject.connectSlotsByName(IssueReporter)
    # setupUi

    def retranslateUi(self, IssueReporter):
        IssueReporter.setWindowTitle(QCoreApplication.translate("IssueReporter", u"Report an Issue", None))
        self.label.setText(QCoreApplication.translate("IssueReporter", u"Consider including the log for the session that had an issue, if applicable.", None))
        self.label_logs_view.setText(QCoreApplication.translate("IssueReporter", u"View Logs:", None))
        self.label_2.setText(QCoreApplication.translate("IssueReporter", u"Log file is at:", None))
        self.pushButton_clipboard.setText(QCoreApplication.translate("IssueReporter", u"Copy to Clipboard", None))
        self.pushButton_open_issues.setText(QCoreApplication.translate("IssueReporter", u"Open Issue Tracker", None))
        self.pushButton_close.setText(QCoreApplication.translate("IssueReporter", u"Close", None))
    # retranslateUi




# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'ImageMatchOverview.ui'
##
## Created by: Qt User Interface Compiler version 6.7.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QAbstractButton, QAbstractItemView, QApplication, QDialog,
    QDialogButtonBox, QFrame, QHBoxLayout, QHeaderView,
    QLabel, QPushButton, QSizePolicy, QSpacerItem,
    QTableWidgetItem, QVBoxLayout, QWidget)

from pcleaner.gui.CustomQ.CTableWidget import CTableWidget

class Ui_ImageMatchOverview(object):
    def setupUi(self, ImageMatchOverview):
        if not ImageMatchOverview.objectName():
            ImageMatchOverview.setObjectName(u"ImageMatchOverview")
        ImageMatchOverview.resize(785, 620)
        self.verticalLayout = QVBoxLayout(ImageMatchOverview)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(1, -1, 1, -1)
        self.verticalLayout_2 = QVBoxLayout()
        self.verticalLayout_2.setSpacing(12)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.verticalLayout_2.setContentsMargins(6, 4, 6, 4)
        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.label_stats = QLabel(ImageMatchOverview)
        self.label_stats.setObjectName(u"label_stats")
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_stats.sizePolicy().hasHeightForWidth())
        self.label_stats.setSizePolicy(sizePolicy)
        self.label_stats.setText(u"<statistics>")

        self.horizontalLayout.addWidget(self.label_stats)

        self.label_warning_icon = QLabel(ImageMatchOverview)
        self.label_warning_icon.setObjectName(u"label_warning_icon")
        self.label_warning_icon.setText(u"<warning icon>")

        self.horizontalLayout.addWidget(self.label_warning_icon)

        self.label_warning = QLabel(ImageMatchOverview)
        self.label_warning.setObjectName(u"label_warning")
        self.label_warning.setText(u"<x pages of OCR data will be lost>")

        self.horizontalLayout.addWidget(self.label_warning)


        self.verticalLayout_2.addLayout(self.horizontalLayout)

        self.horizontalLayout_3 = QHBoxLayout()
        self.horizontalLayout_3.setObjectName(u"horizontalLayout_3")
        self.pushButton_deselect_all = QPushButton(ImageMatchOverview)
        self.pushButton_deselect_all.setObjectName(u"pushButton_deselect_all")

        self.horizontalLayout_3.addWidget(self.pushButton_deselect_all)

        self.pushButton_select_all = QPushButton(ImageMatchOverview)
        self.pushButton_select_all.setObjectName(u"pushButton_select_all")

        self.horizontalLayout_3.addWidget(self.pushButton_select_all)

        self.pushButton_deselect_new = QPushButton(ImageMatchOverview)
        self.pushButton_deselect_new.setObjectName(u"pushButton_deselect_new")
        sizePolicy1 = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Fixed)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.pushButton_deselect_new.sizePolicy().hasHeightForWidth())
        self.pushButton_deselect_new.setSizePolicy(sizePolicy1)

        self.horizontalLayout_3.addWidget(self.pushButton_deselect_new)

        self.horizontalSpacer = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_3.addItem(self.horizontalSpacer)


        self.verticalLayout_2.addLayout(self.horizontalLayout_3)


        self.verticalLayout.addLayout(self.verticalLayout_2)

        self.file_table = CTableWidget(ImageMatchOverview)
        if (self.file_table.columnCount() < 3):
            self.file_table.setColumnCount(3)
        __qtablewidgetitem = QTableWidgetItem()
        __qtablewidgetitem.setText(u"image_index");
        self.file_table.setHorizontalHeaderItem(0, __qtablewidgetitem)
        __qtablewidgetitem1 = QTableWidgetItem()
        self.file_table.setHorizontalHeaderItem(1, __qtablewidgetitem1)
        __qtablewidgetitem2 = QTableWidgetItem()
        self.file_table.setHorizontalHeaderItem(2, __qtablewidgetitem2)
        self.file_table.setObjectName(u"file_table")
        self.file_table.setFocusPolicy(Qt.NoFocus)
        self.file_table.setFrameShape(QFrame.NoFrame)
        self.file_table.setEditTriggers(QAbstractItemView.NoEditTriggers)
        self.file_table.setProperty("showDropIndicator", False)
        self.file_table.setDragDropOverwriteMode(False)
        self.file_table.setDragDropMode(QAbstractItemView.NoDragDrop)
        self.file_table.setAlternatingRowColors(True)
        self.file_table.setSelectionMode(QAbstractItemView.MultiSelection)
        self.file_table.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.file_table.setSortingEnabled(False)
        self.file_table.setCornerButtonEnabled(False)
        self.file_table.horizontalHeader().setVisible(True)
        self.file_table.horizontalHeader().setHighlightSections(False)
        self.file_table.verticalHeader().setVisible(False)
        self.file_table.verticalHeader().setHighlightSections(False)

        self.verticalLayout.addWidget(self.file_table)

        self.horizontalLayout_2 = QHBoxLayout()
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.horizontalLayout_2.setContentsMargins(6, -1, 6, -1)
        self.buttonBox = QDialogButtonBox(ImageMatchOverview)
        self.buttonBox.setObjectName(u"buttonBox")
        self.buttonBox.setOrientation(Qt.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.Cancel|QDialogButtonBox.Ok)

        self.horizontalLayout_2.addWidget(self.buttonBox)


        self.verticalLayout.addLayout(self.horizontalLayout_2)


        self.retranslateUi(ImageMatchOverview)
        self.buttonBox.accepted.connect(ImageMatchOverview.accept)
        self.buttonBox.rejected.connect(ImageMatchOverview.reject)

        QMetaObject.connectSlotsByName(ImageMatchOverview)
    # setupUi

    def retranslateUi(self, ImageMatchOverview):
        ImageMatchOverview.setWindowTitle(QCoreApplication.translate("ImageMatchOverview", u"Image Selection", None))
        self.pushButton_deselect_all.setText(QCoreApplication.translate("ImageMatchOverview", u"Deselect All", None))
        self.pushButton_select_all.setText(QCoreApplication.translate("ImageMatchOverview", u"Select All", None))
        self.pushButton_deselect_new.setText(QCoreApplication.translate("ImageMatchOverview", u"Unselect images without matching OCR data", None))
        ___qtablewidgetitem = self.file_table.horizontalHeaderItem(1)
        ___qtablewidgetitem.setText(QCoreApplication.translate("ImageMatchOverview", u"Image File", None));
        ___qtablewidgetitem1 = self.file_table.horizontalHeaderItem(2)
        ___qtablewidgetitem1.setText(QCoreApplication.translate("ImageMatchOverview", u"OCR Data", u"Can also call this statistics."));
    # retranslateUi




from typing import Generic, TypeVar, Optional
from pathlib import Path

from attrs import frozen, define

import pcleaner.gui.image_file as imf
import pcleaner.output_structures as ost
import pcleaner.structures as st
import pcleaner.config as cfg


T = TypeVar("T")


class Shared(Generic[T]):
    def __init__(self, initial_value: Optional[T] = None) -> None:
        self._container = {"data": initial_value}

    def get(self) -> Optional[T]:
        return self._container["data"]

    def set(self, value: T) -> None:
        self._container["data"] = value

    def is_none(self) -> bool:
        return self._container["data"] is None

        # class OCRModel(Protocol):
        #     def __call__(self, image: Image, **kwargs) -> str:
        # ...


@frozen
class CleaningReviewOptions:
    """
    A compact structure to hold the options for the output review window.
    This is stored in the Mainwindow driver and revisited upon process completion.
    Also stores the needed information to launch the export process after confirming
    the review.
    """

    review_output: ost.Output
    review_mask_outputs: list[ost.Output]
    show_isolated_text: bool
    export_afterwards: bool
    config: cfg.Config
    output_directory: Path
    requested_outputs: list[ost.Output]
    image_files: list[imf.ImageFile]


@define
class OcrReviewOptions:
    """
    A compact structure to hold the options for the OCR review window.
    This one needs to be mutable for the mainwindow to drop off the
    OCR results. The processors isn't run again after this so
    no configuration is needed.
    """

    image_files: list[imf.ImageFile]
    ocr_results: list[st.OCRAnalytic]
    output_path: Path
    csv_output: bool
    editing_old_data: bool



import os
from collections import namedtuple
from pathlib import Path

from PIL import Image
from loguru import logger
from simple_lama_inpainting import SimpleLama

import pcleaner.config as cfg
import pcleaner.image_ops as ops
import pcleaner.structures as st
import pcleaner.model_downloader as md
import pcleaner.output_structures as ost


class InpaintingModel:
    def __init__(self, config: cfg.Config) -> None:
        self.model_path = md.get_inpainting_model_path(config)
        # Sanity check: Make sure the model exists.
        if not self.model_path.is_file():
            raise FileNotFoundError(f"Model not found: {self.model_path}")
        # Load the path into the env variable: LAMA_MODEL
        os.environ["LAMA_MODEL"] = str(self.model_path)
        self.simple_lama = SimpleLama()

    def __call__(self, image: Image, mask: Image) -> Image:
        """
        Inpaint the image using the mask.
        The mask must be a 1-channel image where 1 is the area to be inpainted and 0 is the area to keep.

        :param image: Input image.
        :param mask: Mask image.
        :return: The inpainted image.
        """
        # Run the model but ensure the output image is the same size as the input.
        inpainted_image = self.simple_lama(image, mask)
        if inpainted_image.size != image.size:
            width, height = image.size
            inpainted_image = inpainted_image.crop((0, 0, width, height))
        return inpainted_image


def inpaint_page(i_data: st.InpainterData, model: InpaintingModel) -> Image:
    """
    Load the MaskData from the json file and perform the inpainting process.

    :param i_data: All the data needed for the inpainting process.
    :param model: The inpainting model.
    :return: Analytics.
    """
    # Alias.
    m_conf = i_data.masker_config
    d_conf = i_data.denoiser_config
    i_conf = i_data.inpainter_config

    # Load all the cached data.
    page_data = st.PageData.from_json(i_data.page_data_json_path.read_text(encoding="utf-8"))
    exact_mask_image = Image.open(page_data.mask_path)
    mask_data = st.MaskData.from_json(i_data.mask_data_json_path.read_text(encoding="utf-8"))
    mask_image = Image.open(mask_data.mask_path)

    original_image = Image.open(mask_data.original_path)
    mask_image_bit = mask_image.convert("1")
    exact_mask_image = exact_mask_image.convert("1")
    original_image = original_image.convert("RGB")
    original_path: Path = mask_data.original_path

    path_gen = ost.OutputPathGenerator(original_path, i_data.cache_dir, i_data.page_data_json_path)

    # Collect the boxes to inpaint.
    BoxInpaintData = namedtuple("BoxInpaintData", ["box", "image", "deviation"])
    boxes_to_inpaint: list[BoxInpaintData] = []
    analytics_thicknesses: list[int] = []

    # First, find boxes that failed to be masked.
    failed_boxes_deviation: list[tuple[st.Box, float]] = [
        (box, deviation)
        for box, deviation, failed, thickness in mask_data.boxes_with_stats
        if failed
    ]
    # Next, find the boxes meeting the minimum deviation and maximum thickness.
    poorly_fitted_boxes_deviation: list[tuple[st.Box, float]] = [
        (box, deviation)
        for box, deviation, failed, thickness in mask_data.boxes_with_stats
        if not failed
        and deviation >= i_conf.inpainting_min_std_dev
        and thickness
        is not None  # For box masks, this is none. We don't need to inpaint those, they are always good.
        and thickness <= i_conf.min_inpainting_radius
    ]
    # For the failed boxes, synthesize new masks grown to the minimum size.
    # Sample from the exact mask.
    for box, deviation in failed_boxes_deviation:
        mask = ops.cut_out_box(exact_mask_image, box)
        # Grow the mask to the minimum size.
        mask = ops.grow_mask(mask, m_conf.min_mask_thickness)
        boxes_to_inpaint.append(BoxInpaintData(box, mask, deviation))

    for box, deviation in poorly_fitted_boxes_deviation:
        mask = ops.cut_out_box(mask_image_bit, box)
        boxes_to_inpaint.append(BoxInpaintData(box, mask, deviation))

    # Now grow each mask according to the minimum inpainting radius, and the box's deviation.
    padded_boxes_to_inpaint: list[BoxInpaintData] = []
    for box, mask, deviation in boxes_to_inpaint:
        growth = i_conf.min_inpainting_radius
        growth += int(deviation * i_conf.inpainting_radius_multiplier)
        growth = min(growth, i_conf.max_inpainting_radius)
        # The isolation radius is added ahead of time here. We will grow it by this later, but only after
        # the inpainting.
        growth_with_isolation = growth + i_conf.inpainting_isolation_radius
        # Grow the box, then the mask, ensuring it stays within the image bounds.
        box_padded = box.pad(growth_with_isolation, mask_image.size)
        mask_padded = Image.new("1", mask_image.size, 0)
        offset: tuple[int, int] = box.x1 - box_padded.x1, box.y1 - box_padded.y1
        mask_padded.paste(mask, offset)
        mask_padded = ops.grow_mask(mask_padded, growth)
        padded_boxes_to_inpaint.append(BoxInpaintData(box_padded, mask_padded, deviation))
        analytics_thicknesses.append(growth)

    # Merge all the masks into one, then scale it up to the original image size.
    combined_mask = Image.new("1", mask_image.size, 0)
    for box, mask, _ in padded_boxes_to_inpaint:
        combined_mask.paste(mask, (box.x1, box.y1), mask)

    combined_mask.save(path_gen.inpainting_mask)

    # Scale up the masks before inpainting.
    if original_image.size != mask_image.size:
        combined_mask = combined_mask.resize(original_image.size, resample=Image.NEAREST)

    # Inpaint the image.
    if boxes_to_inpaint:
        inpainted_image = model(original_image, combined_mask)
    else:
        inpainted_image = original_image

    # Lastly, grow the masks again by the isolation radius to cut out the inpainted areas.
    isolated_combined_mask = Image.new("1", mask_image.size, 0)
    for box, mask, deviation in padded_boxes_to_inpaint:
        mask = ops.grow_mask(mask, i_conf.inpainting_isolation_radius)
        isolated_combined_mask.paste(mask, (box.x1, box.y1), mask)
    if original_image.size != mask_image.size:
        isolated_combined_mask = isolated_combined_mask.resize(
            original_image.size, resample=Image.NEAREST
        )

    # Apply the mask to the inpainted image.
    inpainted_areas = Image.new("RGBA", original_image.size, 0)
    inpainted_areas.paste(inpainted_image, (0, 0), isolated_combined_mask)

    # Create a new output with these inpainted areas overlayed.
    # But first, apply the cleaning masks.
    cleaned_image = (
        original_image  # Don't bother copying as we won't need this anymore, so overwrite.
    )
    # We need to scale up the mask image to the original image size.
    if original_image.size != mask_image.size:
        mask_image = mask_image.resize(original_image.size, resample=Image.NEAREST)
    cleaned_image.paste(mask_image, (0, 0), mask_image)
    # Then, if denoising was enabled, apply that next.
    if d_conf.denoising_enabled and path_gen.noise_mask.is_file():
        noise_mask = Image.open(path_gen.noise_mask)
        cleaned_image.paste(noise_mask, (0, 0), noise_mask)

    cleaned_image.paste(inpainted_areas, (0, 0), isolated_combined_mask)

    # Save output.
    inpainted_areas.save(path_gen.inpainting)
    cleaned_image.save(path_gen.clean_inpaint)

    # Package the analytics. We're only interested in the thicknesses.
    return st.InpaintingAnalytic(tuple(analytics_thicknesses), original_path)
